import json
import logging
import math
import time, os
from pathlib import Path
import numpy as np

from openmines.src.utils.event import EventPool
from openmines.src.utils.logger import MineLogger
from functools import reduce

# 项目主路径
PROJECT_ROOT_PATH = Path(__file__).parent.parent.parent
LOG_FILE_PATH = PROJECT_ROOT_PATH / "src" / "data" / "logs"

# 车辆状态
ON_CHARGING_SITE = -1  # 在充电站
ON_ROAD_TO_FIRST_LOAD = -2  # 从充电站到第一个装载点
UNHAULING = 0  # 从卸载区到装载区
WAIT_FOR_LOAD = 1  # 等待装载
LOADING = 2  # 装载中
HAULING = 3  # 从装载区到卸载区
WAIT_FOR_DUMP = 4  # 等待卸载
DUMPING = 5  # 卸载中
REPAIRING = 6  # 维修中
UNREPAIRABLE = 7  # 无法维修，返回充电场

class TickGenerator:
    """
    读取矿山下每个Truck的EventPool进行分析统计
    每一分钟相当于一个tick
    """

    def __init__(self, mine:"Mine", tick_num=60*8):
        self.mine = mine
        self.tick_num = tick_num
        self.ticks = dict()
        self.result_path = os.getcwd() + '/results'
        if not os.path.exists(self.result_path):
            os.makedirs(self.result_path)


    def run(self):
        """
        获取矿山下每个Truck的EventPool并进行总体分析
        1.获取需要dump的对象
        2.在每个tick中按照对象类别分别进行统计，形成字典格式
        :return:
        """
        # 读取mine的配置文件 获取 装载点(铲车、停车场)、卸载点（卸载位、停车场）、充电区 在图像中的位置
        load_sites = self.mine.load_sites
        dump_sites = self.mine.dump_sites
        charging_site = self.mine.charging_site
        trucks = self.mine.trucks

        # 遍历产生每个tick
        for tick in range(self.tick_num):
            # print(f"tick: {tick}")
            cur_time = tick
            truck_states = dict()
            load_site_states = dict()
            dump_site_states = dict()
            shovel_states = dict()
            dumper_states = dict()

            for truck in self.mine.trucks:
                event_pool = truck.event_pool
                # 判断卡车状态
                past_events = event_pool.get_event_by_time(cur_time+0.1)
                future_events = event_pool.get_event_by_time(cur_time+0.1, mode="future")
                status = None
                position = None
                truck_position = np.array([0.0, 0.0])
                ## 车辆在充电场
                if len(past_events)==0:
                    status = -1  # 没有事件、仅有一个调度事件==还在空载状态，位置处于充电站
                    truck_position = np.array(charging_site.position) + np.array([0.0, 0.0])
                ## 车辆在从充电场到装载点的路上
                if len(past_events)==2 and past_events[-1].event_type=="init":
                    """init example:
                        info={"name": self.name, "status": event_name, "speed": manual_speed if manual_speed is not None else self.truck_speed,
                                              "start_time": self.env.now, "est_end_time": self.env.now+duration, "end_time": None,
                                              "start_location": self.current_location.name,
                                              "target_location": target_location.name,
                                              "distance": distance, "duration": None}
                    """
                    status = -2  # 卡车已经得到了订单并且已经开始从充电站移动到第一个装载点

                    # 计算卡车位置
                    last_event = past_events[-1]
                    # 时间计算
                    cur_time_delta = cur_time - last_event.info["start_time"]
                    time_ratio = cur_time_delta / (last_event.info["est_end_time"] - last_event.info["start_time"])
                    # 方向计算
                    target_load_site = self.mine.get_dest_obj_by_name(last_event.info["target_location"])
                    direction = np.array(target_load_site.position) - np.array(charging_site.position)
                    truck_position = np.array(charging_site.position) + time_ratio * direction

                ## 车辆在装载点等待装载
                if past_events[-1].event_type=="wait shovel":
                    """wait shovel example:
                    info={"name": self.name, "status": "waiting for shovel",
                                            "start_time": self.env.now, "end_time": None,
                                            "shovel": shovel.name, "wait_duration": None}
                    """
                    status = 1
                    # 计算卡车位置
                    # 等待装载位置=装载区位置+停车场相对位置偏移=停车场位置
                    truck_cur_location_name = truck.current_location.name
                    cur_load_site = self.mine.get_dest_obj_by_name(truck_cur_location_name)
                    parkinglot = cur_load_site.parking_lot

                    queue_index = past_events[-1].info["queue_index"]
                    shovel_name = past_events[-1].info["shovel"]
                    shovel = self.mine.get_service_vehicle_by_name(shovel_name)
                    truck_position = np.array(shovel.position) - np.array([0.05, 0.0]) - (queue_index+1)*np.array([0.035,0.0])  # 已经处理过

                ## 车辆在装载点装载
                if past_events[-1].event_type=="get shovel":
                    """get shovel example:
                    info={"name": self.name, "status": "loading on shovel",
                                                "start_time": self.env.now, "end_time": None,
                                                "shovel": shovel.name, "load_duration": None}
                    """
                    status = 2
                    # 计算卡车位置
                    # 装载位置=装载区位置+铲子相对位置偏移=铲子位置
                    shovel = self.mine.get_service_vehicle_by_name(past_events[-1].info["shovel"])
                    loading_position = np.array(shovel.position) - np.array([0.05, 0.0])
                    truck_position = loading_position

                ## 车辆满载 从装载点前往卸载点中
                if past_events[-1].event_type=="haul":
                    """haul example:
                    info={"name": self.name, "status": event_name, "speed": manual_speed if manual_speed is not None else self.truck_speed,
                                              "start_time": self.env.now, "est_end_time": self.env.now+duration, "end_time": None,
                                              "start_location": self.current_location.name,
                                              "target_location": target_location.name,
                                              "distance": distance, "duration": None}
                    """
                    status = 3

                    # 计算卡车位置
                    # 时间计算
                    last_event = past_events[-1]
                    cur_time_delta = cur_time - last_event.info["start_time"]
                    end_time = last_event.info["end_time"] if last_event.info["end_time"] is not None else last_event.info["est_end_time"]
                    time_ratio = cur_time_delta / (end_time - last_event.info["start_time"])
                    # 方向计算
                    cur_load_site = self.mine.get_dest_obj_by_name(last_event.info["start_location"])
                    target_unload_site = self.mine.get_dest_obj_by_name(last_event.info["target_location"])
                    direction = np.array(target_unload_site.position) - np.array(cur_load_site.position)
                    truck_position = np.array(cur_load_site.position) + time_ratio * direction

                ## 车辆在卸载点等待卸载
                if past_events[-1].event_type=="wait dumper":
                    """wait dumper example:
                    info={"name": self.name, "status": "waiting for dumper",
                                                    "start_time": self.env.now, "end_time": None,
                                                    "dumper": dumper.name, "wait_duration": None}
                    
                    """
                    status = 4
                    # 计算卡车位置
                    # 等待卸载位置=卸载区位置+停车场相对位置偏移=停车场位置
                    truck_cur_location_name = truck.current_location.name
                    parkinglot = self.mine.get_dest_obj_by_name(truck_cur_location_name).parking_lot
                    truck_position = np.array(parkinglot.position)  # 已经处理过

                ## 车辆在卸载点卸载
                if past_events[-1].event_type=="get dumper":
                    """get dumper example:
                    info={"name": self.name, "status": "unloading on dumper",
                                                    "start_time": self.env.now, "end_time": None,
                                                    "dumper": dumper.name, "unload_duration": None}
                    """
                    status = 5
                    # 计算卡车位置
                    # 卸载位置=卸载区位置+dumper相对位置偏移=dumper位置
                    dumper = self.mine.get_service_vehicle_by_name(past_events[-1].info["dumper"])
                    truck_position = np.array(dumper.position)  # 已经处理过

                ## 车辆空载 从卸载点前往装载点中
                if past_events[-1].event_type=="unhaul":
                    """unhaul example:
                    info={"name": self.name, "status": event_name, "speed": manual_speed if manual_speed is not None else self.truck_speed,
                                              "start_time": self.env.now, "est_end_time": self.env.now+duration, "end_time": None,
                                              "start_location": self.current_location.name,
                                              "target_location": target_location.name,
                                              "distance": distance, "duration": None}
                    """
                    status = 0
                    # 计算卡车位置
                    # 时间计算
                    last_event = past_events[-1]
                    cur_time_delta = cur_time - last_event.info["start_time"]
                    end_time = last_event.info["end_time"] if last_event.info["end_time"] is not None else last_event.info["est_end_time"]
                    time_ratio = cur_time_delta / (end_time - last_event.info["start_time"])
                    # 方向计算
                    cur_unload_site = self.mine.get_dest_obj_by_name(last_event.info["start_location"])
                    target_load_site = self.mine.get_dest_obj_by_name(last_event.info["target_location"])
                    direction = np.array(target_load_site.position) - np.array(cur_unload_site.position)
                    truck_position = np.array(cur_unload_site.position) + time_ratio * direction

                # 车辆损坏事件，则车辆在原地不动
                if "breakdown" in past_events[-1].event_type:
                    status = 6
                    # 计算卡车位置
                    # 车辆损坏后位置=上一时间位置
                    truck_position = truck_position  # 已经处理过

                if "unrepairable" in past_events[-1].event_type:
                    status = 7
                    # 计算卡车位置
                    # 车辆完全损坏后位置=充电区位置
                    # 获取chargingSite位置
                    truck_position = np.array(self.mine.charging_site.position).astype(float)  # 已经处理过

                truck_state = {
                    "name":truck.name,
                    "time":cur_time,
                    "state":status,
                    "position":list(truck_position if truck_position is not None else (0.5,0.5)),
                }
                truck_states[truck.name] = truck_state

            # 统计每一个loadsite的信息
            for loadsite in self.mine.load_sites:
                # 统计停车场的信息
                ## 获得停车场的排队数量
                total_queue_length = loadsite.parking_lot.queue_status["total"].get(cur_time,0)
                ## 统计loadsite的信息
                loadsite_status = loadsite.status.get(cur_time,{"produced_tons":0,"service_count":0})
                load_site_tick = {
                    "name":loadsite.name,
                    "tons":loadsite_status["produced_tons"],
                    "service_count":loadsite_status["service_count"],
                    "time":cur_time,
                    "total_queue_length":total_queue_length,
                    "shovel_queue_length":{},
                    "position":list(loadsite.position)
                }

                # 统计shovel的信息
                for shovel in loadsite.shovel_list:
                    # 获得shovel的排队数量
                    shovel_queue_length = loadsite.parking_lot.queue_status[shovel.name].get(cur_time,0)
                    load_site_tick["shovel_queue_length"][shovel.name] = shovel_queue_length
                    # 统计shovel的信息
                    shovel_cur_state = shovel.status.get(cur_time,{"produced_tons":0,"service_count":0})
                    shovel_tick = {
                        "name":shovel.name,
                        "tons":shovel_cur_state["produced_tons"],
                        "service_count":shovel_cur_state["service_count"],
                        "time":cur_time,
                        "queue_length":shovel_queue_length,
                        "position":(shovel.position),
                    }
                    shovel_states[shovel.name] = shovel_tick
                load_site_states[loadsite.name] = load_site_tick

            # 统计每一个dumpsite的信息
            for dumpsite in self.mine.dump_sites:
                # 统计停车场的信息
                ## 获得停车场的排队数量
                total_queue_length = dumpsite.parking_lot.queue_status["total"].get(cur_time,0)
                # 统计dumpsite的信息
                dump_site_status = dumpsite.status.get(cur_time,{"produced_tons":0,"service_count":0})
                dump_site_tick = {
                    "name":dumpsite.name,
                    "tons":dump_site_status["produced_tons"],
                    "service_count":dump_site_status["service_count"],
                    "time":cur_time,
                    "total_queue_length":total_queue_length,
                    "dumper_queue_length":{},
                    "position":list(dumpsite.position)
                }

                # 统计dumper的信息
                for dumper in dumpsite.dumper_list:
                    # 获得dumper的排队数量
                    dumper_queue_length = dumpsite.parking_lot.queue_status[dumper.name].get(cur_time,0)
                    dump_site_tick["dumper_queue_length"][dumper.name] = dumper_queue_length
                    # 统计dumper的信息
                    dumper_cur_state = dumper.status.get(cur_time,{"produced_tons":0,"service_count":0})

                    dumper_tick = {
                        "name":dumper.name,
                        "tons":dumper_cur_state["produced_tons"],
                        "service_count":dumper_cur_state["service_count"],
                        "time":cur_time,
                        "queue_length":dumper_queue_length,
                        "position":list(dumper.position),
                    }
                    dumper_states[dumper.name] = dumper_tick
                dump_site_states[dumpsite.name] = dump_site_tick

            # 统计矿山整体信息
            mine_status = self.mine.status.get(cur_time,{"produced_tons":0,"service_count":0,"working_truck_count":0,"waiting_truck_count":0,"load_unload_truck_count":0,"moving_truck_count":0,"repairing_truck_count":0,"road_jam_count":0,"road_repair_count":0,"truck_repair":0,"truck_unrepairable":0,"random_event_count":0})
            mine_tick = {
                "time":cur_time,
                # KPIs
                "produced_tons":mine_status["produced_tons"],
                "service_count":mine_status["service_count"],
                # stats
                "working_truck_count": mine_status["working_truck_count"],
                "waiting_truck_count": mine_status["waiting_truck_count"],
                "load_unload_truck_count": mine_status["load_unload_truck_count"],
                "moving_truck_count": mine_status["moving_truck_count"],
                "repairing_truck_count": mine_status["repairing_truck_count"],
                # event stats
                "road_jam_count": mine_status["road_jam_count"],
                "road_repair_count": mine_status["road_repair_count"],
                "truck_repair": mine_status["truck_repair"],
                "truck_unrepairable": mine_status["truck_unrepairable"],
                "random_event_count":mine_status["random_event_count"]
            }


            tick_packet = {
                "tick":cur_time,
                "truck_states":truck_states,
                "load_site_states":load_site_states,
                "dump_site_states":dump_site_states,
                "shovel_states":shovel_states,
                "dumper_states":dumper_states,
                "mine_states":mine_tick
            }
            self.ticks[cur_time] = tick_packet

        # 添加一些分析性非实时性的统计数据
        # 统计卸载区产量
        print(f"{self.mine.name} summary: {self.ticks[cur_time]['mine_states']['produced_tons']} tons")
        # 统计MatchingFactor
        # paper:Match factor for heterogeneous truck and loader fleets
        shovels = [shovel for load_site in self.mine.load_sites for shovel in load_site.shovel_list]
        trucks = self.mine.trucks
        truck_cycle_time_avg = np.mean([truck.truck_cycle_time for truck in trucks])

        num_trucks = len(self.mine.trucks)
        num_shovels = len(shovels)
        loading_time:np.array = np.zeros((num_trucks,num_shovels))
        for i,truck in enumerate(trucks):
            for j,shovel in enumerate(shovels):
                loading_time[i, j] = round((truck.truck_capacity / shovel.shovel_tons), 1) * shovel.shovel_cycle_time# in mins

        # 按行按列去重复
        loading_time = np.unique(loading_time, axis=0)
        unique_loading_time = np.unique(loading_time, axis=1).astype(int)
        # 对异构铲车的子构数量进行统计
        shovel_type_count = dict()
        for i in range(unique_loading_time.shape[0]):  # truck type index
                int_data = np.array(loading_time[i,:]).astype(int)
                for value in set(int_data):
                    shovel_type_count[f'{i}_{value}'] = list(int_data).count(value)  # it means truck type i w.r.t shovel type num

        # unique_loading_time = np.ones_like(unique_loading_time) + unique_loading_time
        # 对每一行求lcm
        lcm_load_time = np.lcm.reduce(unique_loading_time, axis=1)
        upside_down_sum = 0
        for i in range(unique_loading_time.shape[0]):
            for j in range(unique_loading_time.shape[1]):
                upside_down_sum += shovel_type_count[f'{i}_{unique_loading_time[i,j]}']*(lcm_load_time[i]/unique_loading_time[i,j])
        match_factor = (num_trucks*np.sum(lcm_load_time)) / (upside_down_sum*truck_cycle_time_avg)
        print(f'MatchingFactor: {match_factor}')
        total_wait_time = sum([truck.get_wait_time() for truck in self.mine.trucks])
        print(f'TotalWaitTime: {total_wait_time}')
        # 统计调度算法代码的执行性能
        total_order_count = self.mine.dispatcher.total_order_count
        total_order_time = self.mine.dispatcher.total_order_time
        avg_time_per_order = total_order_time / total_order_count if total_order_count > 0 else 0

        self.ticks["summary"] = {
            "produced_tons":self.ticks[cur_time]['mine_states']['produced_tons'],
            "MatchingFactor":match_factor,
            "TotalWaitTime":total_wait_time,
            "avg_time_per_order":avg_time_per_order,
            "total_order_count":total_order_count,
            "RoadJams":self.ticks[cur_time]['mine_states']['road_jam_count']
        }


    def write_to_file(self, file_name):
        """
        将ticks写入文件
        :param file_path:
        :return:
        """
        file_path = os.path.join(self.result_path, file_name)
        with open(file_path, "w") as f:
            try:
                json.dump(self.ticks,f)
                print("file_name:{} write success".format(file_name))
            except Exception as e:
                print(e)
                print("file_name:{} write failed".format(file_name))
        return self.ticks

    def read_from_file(self, file_name):
        """
        从文件中读取ticks
        :param file_path:
        :return:
        """
        file_path = os.path.join(self.result_path, file_name)
        with open(file_path, "r") as f:
            try:
                self.ticks = json.load(f)
                print("file_name:{} read success".format(file_name))
            except Exception as e:
                print(e)
                print("file_name:{} read failed".format(file_name))


from __future__ import annotations

import json

import numpy as np

from openmines.src.charging_site import ChargingSite
from openmines.src.dispatcher import BaseDispatcher
from openmines.src.load_site import LoadSite, Shovel
from openmines.src.dump_site import DumpSite, Dumper
from openmines.src.road import Road
from openmines.src.truck import Truck


class FixedGroupDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "FixedGroupDispatcher"
        self.group_solution = None  # Solution will be stored here after computation

    def compute_solution(self, mine: "Mine"):
        if self.group_solution is not None:
            return  # Solution already computed

        self.group_solution = {}  # Initialize group_solution
        load_sites = mine.load_sites
        dump_sites = mine.dump_sites

        # Compute load site productivity
        for load_site in load_sites:
            load_site_productivity = sum(
                shovel.shovel_tons / shovel.shovel_cycle_time for shovel in load_site.shovel_list)
            load_site.load_site_productivity = load_site_productivity

        total_load_site_productivity = sum(load_site.load_site_productivity for load_site in load_sites)

        # Calculate load site capacity
        trucks = mine.trucks
        total_truck_capacity = sum(truck.truck_capacity for truck in trucks)
        load_site_capacity = {
            load_site.name: (load_site.load_site_productivity / total_load_site_productivity) * total_truck_capacity
            for load_site in load_sites
        }

        # Initialize the group solution structure
        for load_site in load_sites:
            self.group_solution[load_site.name] = {
                "trucks": [],
                "dumpsite": None
            }

        # Distribute trucks to load sites
        for truck in trucks:
            for load_site in load_sites:
                if sum(t.truck_capacity for t in self.group_solution[load_site.name]["trucks"]) < load_site_capacity[
                    load_site.name]:
                    self.group_solution[load_site.name]["trucks"].append(truck)
                    break

        # Match load sites with the nearest dump site
        for load_site in load_sites:
            distances = [np.linalg.norm(np.array(load_site.position) - np.array(dump_site.position)) for dump_site in
                         dump_sites]
            nearest_dump_site = dump_sites[distances.index(min(distances))]
            self.group_solution[load_site.name]["dumpsite"] = nearest_dump_site

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        if self.group_solution is None:
            self.compute_solution(mine)

        for load_site_name, info in self.group_solution.items():
            if truck in info["trucks"]:
                return mine.load_sites.index([ls for ls in mine.load_sites if ls.name == load_site_name][0])

        return 0

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        """
        前往最近距离的卸载点卸载
        :param truck:
        :param mine:
        :return:
        """
        current_location = truck.current_location
        assert isinstance(current_location, LoadSite), "current_location is not a LoadSite"
        cur_index = mine.load_sites.index(current_location)
        cur_to_dump = mine.road.road_matrix[cur_index, :]
        min_index = cur_to_dump.argmin()
        return min_index

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        """
        查询当前车辆所在的装载点并返回
        :param truck:
        :param mine:
        :return:
        """
        for load_site_name, info in self.group_solution.items():
            if truck in info["trucks"]:
                return mine.load_sites.index([ls for ls in mine.load_sites if ls.name == load_site_name][0])

        return 0  # This can be expanded based on specific logic

# Test the dispatcher
if __name__ == "__main__":
    dispatcher = FixedGroupDispatcher()
    config_file = "/Users/mac/PycharmProjects/truck_shovel_mix/openmines_project/openmines/src/conf/north_pit_mine.json"
    from openmines.src.mine import Mine

    def load_config(filename):
        with open(filename, 'r') as file:
            return json.load(file)

    config = load_config(config_file)
    mine = Mine(config_file)
    # 初始化充电站和卡车
    charging_site = ChargingSite(config['charging_site']['name'], position=config['charging_site']['position'])
    for truck_config in config['charging_site']['trucks']:
        for _ in range(truck_config['count']):
            truck = Truck(
                name=f"{truck_config['type']}{_ + 1}",
                truck_capacity=truck_config['capacity'],
                truck_speed=truck_config['speed']
            )
            charging_site.add_truck(truck)

    # 初始化装载点和铲车
    for load_site_config in config['load_sites']:
        load_site = LoadSite(name=load_site_config['name'], position=load_site_config['position'])
        for shovel_config in load_site_config['shovels']:
            shovel = Shovel(
                name=shovel_config['name'],
                shovel_tons=shovel_config['tons'],
                shovel_cycle_time=shovel_config['cycle_time'],
                position_offset=shovel_config['position_offset']
            )
            load_site.add_shovel(shovel)
        load_site.add_parkinglot(position_offset=load_site_config['parkinglot']['position_offset'],
                                 name=load_site_config['parkinglot']['name'])
        mine.add_load_site(load_site)

    # 初始化卸载点和卸载机
    for dump_site_config in config['dump_sites']:
        dump_site = DumpSite(dump_site_config['name'], position=dump_site_config['position'])
        for dumper_config in dump_site_config['dumpers']:
            for _ in range(dumper_config['count']):
                dumper = Dumper(
                    name=f"{dump_site_config['name']}-点位{_}",
                    dumper_cycle_time=dumper_config['cycle_time'],
                    position_offset=dumper_config['position_offset']
                )
                dump_site.add_dumper(dumper)
        dump_site.add_parkinglot(position_offset=dump_site_config['parkinglot']['position_offset'],
                                 name=dump_site_config['parkinglot']['name'])
        mine.add_dump_site(dump_site)

    # 初始化道路
    road_matrix = np.array(config['road']['road_matrix'])
    road_event_params = config['road'].get('road_event_params', {})  # 从配置中加载道路事件参数

    charging_to_load_road_matrix = config['road']['charging_to_load_road_matrix']
    road = Road(road_matrix=road_matrix, charging_to_load_road_matrix=charging_to_load_road_matrix,
                road_event_params=road_event_params)
    # # 添加充电站和装载区卸载区
    mine.add_road(road)
    mine.add_charging_site(charging_site)
    ## 添加调度
    mine.add_dispatcher(dispatcher)
    dispatcher.compute_solution(mine)
    print(dispatcher.group_solution)
    # Assume mine and trucks are defined somewhere
    # print(dispatcher.give_init_order(truck, mine))
    # print(dispatcher.give_haul_order(truck, mine))
    # print(dispatcher.give_back_order(truck, mine))



from __future__ import annotations
import numpy as np  # 导入NumPy库
import random,json,time
import openai


from openmines.src.dispatcher import BaseDispatcher
from openmines.src.load_site import LoadSite, Shovel
from openmines.src.dump_site import DumpSite, Dumper


class LLMDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "LLMDispatcher"
        self.OPENAI = OPENAI(model_name="gpt-3.5-turbo-0613")
        self.order_index = 0
        self.init_order_index = 0
        self.common_order_index = 0
        self.init_order_history = []
        self.haul_order_history = []
        self.back_order_history = []
        self.order_history = []
        self.np_random = np.random.RandomState()  # 创建NumPy的随机状态对象

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        # logger
        self.logger = mine.global_logger.get_logger("LLMDispatcher")
        cur_location = mine.charging_site.name
        # 统计loadsite信息
        load_sites = mine.load_sites
        loadsite_queue_length = [load_site.parking_lot.queue_status["total"][int(mine.env.now)] for load_site in load_sites]
        estimated_loadsite_queue_wait_times = [load_site.estimated_queue_wait_time for load_site in load_sites]

        # 获取dumpsite信息
        avaliable_dumpsites = [dumpsite for dumpsite in mine.dump_sites if dumpsite.parking_lot is not None]
        dump_site_names = [dumpsite.name for dumpsite in avaliable_dumpsites]
        dumpsite_queue_length = [dumpsite.parking_lot.queue_status["total"][int(mine.env.now)] for dumpsite in
                                 avaliable_dumpsites]
        estimated_dumpsite_queue_wait_times = [dumpsite.estimated_queue_wait_time for dumpsite in avaliable_dumpsites]

        # 获取过去的订单信息
        past_orders_all = self.order_history[-10:]
        past_orders_haul = [order for order in self.order_history if order["order_type"] == "haul_order"][-10:]
        # 获取Road距离信息
        road_matrix = mine.road.road_matrix

        prompt = f"""
                You are now an LLM (Large Language Model) scheduler, and your task is to assign an initial task destination for the current truck based on the available information.
                The current truck is in the charging area and needs to go to the loading area for loading.
                Background knowledge:
                    The mine has multiple loading and unloading areas as well as traffic roads. Mining trucks need to transport goods back and forth between these areas.
                    Loading areas have different loading capacities and queue situations, and unloading areas have different unloading capacities and queue situations.
                    The mining trucks are heterogeneous, varying in their running speeds and loading tonnage.
                    If a road has a large number of mining trucks dispatched, the probability of random events such as traffic jams and road repairs increases, leading to longer operation times for the trucks.

                Current mine information:
                        Loading areas: {[{"name": loadsite.name, "type": "loadsite", "load_capability(tons/min)": loadsite.load_site_productivity, "distance": mine.road.charging_to_load[i],
                                          "queue_length": loadsite_queue_length[i],
                                          "estimated_queue_wait_times": estimated_loadsite_queue_wait_times[i]
                                          } for i, loadsite in enumerate(mine.load_sites)]},

                        Current road information:
                            {[{"road_id": f"{cur_location} to {load_site.name}", "road_desc": f"from {cur_location} to {load_site.name}", "distance": mine.road.charging_to_load[j],
                               "trucks_on_this_road": mine.road.road_status[(cur_location, load_site.name)]["truck_count"],
                               "jammed_trucks_on_this_road": mine.road.road_status[(cur_location, load_site.name)]["truck_jam_count"],
                               "is_road_in_repair": mine.road.road_status[(cur_location, load_site.name)]["repair_count"]} for j, load_site in enumerate(mine.load_sites)]}


                Current order information:
                {{
                "cur_time": {mine.env.now},
                "order_type": "init_order",
                "truck_name": "{truck.name}",
                "truck_capacity": {truck.truck_capacity},
                "truck_speed": {truck.truck_speed}
                }}
                Historical scheduling decisions:
                {[{key: val for key, val in order.items() if key not in ['prompt', 'response']} for order in past_orders_all]}


                Please assign an initial task destination for the current truck based on the above information,
                Requirements:
                1. Overall objective: Considering the impact of random events on the road, choose the loading area with the shortest distance as much as possible, while avoiding traffic jams and road repairs.
                2. Finally, based on the overall objective, directly provide the following JSON string as the decision result:
                {{
                    "truck_name": "{truck.name}",
                    "loadingsite_index": an integer from 0 to {len(mine.load_sites) - 1}
                }}

                """

        for i in range(3):
            try:
                response = self.OPENAI.get_response(prompt=prompt)
                self.logger.info(f"LLM 订单{self.order_index +1 }：prompt:{prompt} \n {response}")
                start = response.find('{')
                end = response.rfind('}') + 1
                # 提取 JSON 字符串
                json_str = response[start:end]
                data = json.loads(json_str)
                loadsite_index = data["loadingsite_index"]
                break
            except Exception as e:
                print(e)
                loadsite_index = random.randint(0, len(mine.load_sites) - 1)
                self.logger.error(f"LLM 订单{self.order_index +1 }：parse error，giving random order")


        # logging
        order = {
            "cur_time": mine.env.now,
            "order_type": "init_order",
            "truck_name": truck.name,
            "truck_capacity": truck.truck_capacity,
            "truck_speed": truck.truck_speed,
            "loadingsite_index": loadsite_index,
            "prompt": prompt,
            "response": response}
        self.init_order_history.append(order)
        self.order_history.append(order)
        self.order_index += 1
        self.init_order_index += 1
        self.logger.debug(f"LLM INIT 订单{self.init_order_index}：{order}")
        return loadsite_index

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        # logger
        self.logger = mine.global_logger.get_logger("LLMDispatcher")
        # 获取当前卡车信息
        truck_load = truck.truck_load
        cur_location = truck.current_location.name
        cur_loadsite = mine.get_dest_obj_by_name(cur_location)
        cur_loadsite_index = mine.load_sites.index(cur_loadsite)
        assert isinstance(cur_loadsite, LoadSite), f"the truck {truck.name} is not in a loadsite, it is in {cur_loadsite.name}"
        # 统计loadsite信息
        load_sites = mine.load_sites
        loadsite_queue_length = [load_site.parking_lot.queue_status["total"][int(mine.env.now)] for load_site in load_sites]
        estimated_loadsite_queue_wait_times = [load_site.estimated_queue_wait_time for load_site in load_sites]

        # 获取dumpsite信息
        avaliable_dumpsites = [dumpsite for dumpsite in mine.dump_sites if dumpsite.parking_lot is not None]
        dump_site_names = [dumpsite.name for dumpsite in avaliable_dumpsites]
        dumpsite_queue_length = [dumpsite.parking_lot.queue_status["total"][int(mine.env.now)] for dumpsite in avaliable_dumpsites]
        estimated_dumpsite_queue_wait_times = [dumpsite.estimated_queue_wait_time for dumpsite in avaliable_dumpsites]

        # 获取过去的订单信息
        # past_orders_all = self.order_history[-10:]
        past_orders_haul = [order for order in self.order_history if order["order_type"] == "haul_order"][-10:]
        # 获取Road距离信息
        road_matrix = mine.road.road_matrix

        prompt = f"""
                You are now an LLM scheduler, and your task is to assign a target location for the current truck based on the available information.
                Background knowledge:
                The mine has multiple loading and unloading areas as well as traffic roads, and mining trucks need to transport goods back and forth between these areas.
                The loading capacity and queue situation of each loading area are different, as are the unloading capacity and queue situation of each unloading area.
                The mining trucks are heterogeneous, with differences in their running speeds and loading capacities.
                If a road has many mining trucks, the probability of random events such as traffic jams and road maintenance increases, leading to longer operation times for the trucks.

                Current mine information:
                Loading areas: {[{"name": loadsite.name, "type": "loadsite", "load_capability(tons/min)": loadsite.load_site_productivity, "queue_length": loadsite_queue_length[i]
                                  } for i, loadsite in enumerate(mine.load_sites)]},
                Unloading areas: {[{"name": dumpsite.name, "type": "dumpsite", "index": j, "distance": road_matrix[cur_loadsite_index][j],
                                    "queue_length": dumpsite_queue_length[j]
                                    } for j, dumpsite in enumerate(avaliable_dumpsites)]},
                Current road information:
                    {[{"road_id": f"{cur_location} to {dumpsite.name}", "road_desc": f"from {cur_location} to {dumpsite.name}", "distance": road_matrix[cur_loadsite_index][j],
                       "trucks_on_this_road": mine.road.road_status[(cur_location, dumpsite.name)]["truck_count"],
                       "jammed_trucks_on_this_road": mine.road.road_status[(cur_location, dumpsite.name)]["truck_jam_count"],
                       "is_road_in_repair": mine.road.road_status[(cur_location, dumpsite.name)]["repair_count"]} for j, dumpsite in enumerate(avaliable_dumpsites)]}

                Historical scheduling decisions:
                    {[{key: val for key, val in order.items() if key not in ['prompt', 'response']} for order in past_orders_haul]}

                The current truck is at the loading area {cur_location} and needs to go to the unloading area for unloading.
                Current truck order request information:
                {{
                "cur_time": {mine.env.now},
                "order_type": "haul_order",
                "truck_location": "{truck.current_location.name}",
                "truck_name": "{truck.name}",
                "truck_capacity": {truck.truck_capacity},
                "truck_speed": {truck.truck_speed}
                }}

                Please assign a suitable unloading area as the target location for the current truck based on the information above, allowing it to reach the target location as quickly as possible for unloading:
                Requirements:
                1. Overall objective: Considering the impact of random events on the road, choose the unloading area with the shortest distance as much as possible, while avoiding traffic jams and road repairs.
                2. Finally, based on the overall objective, directly provide the following JSON string as the decision result:
                {{
                    "truck_name": "{truck.name}",
                    "dumpsite_index": an integer from 0 to {len(avaliable_dumpsites) - 1}
                }}

                """

        for i in range(3):
            try:
                response = self.OPENAI.get_response(prompt)
                self.logger.info(f"LLM 订单{self.order_index + 1}：prompt:{prompt} \n {response}")
                start = response.find('{')
                end = response.rfind('}') + 1
                # 提取 JSON 字符串
                json_str = response[start:end]
                data = json.loads(json_str)
                dumpsite_index = data["dumpsite_index"]
                break
            except Exception as e:
                print(e)
                dumpsite_index = random.randint(0, len(avaliable_dumpsites) - 1)
                self.logger.error(f"LLM 订单{self.order_index + 1}：parse error，giving random order")

        # logging
        order = {
            "cur_time": mine.env.now,
            "order_type": "haul_order",
            "truck_name": truck.name,
            "truck_capacity": truck.truck_capacity,
            "truck_speed": truck.truck_speed,
            "truck_location": f"{truck.current_location.name}",
            "dumpsite_index": dumpsite_index,
            "prompt": prompt,
            "response": response}
        self.haul_order_history.append(order)
        self.order_history.append(order)
        self.order_index += 1
        self.logger.debug(f"LLM HAUL 订单{self.order_index}：{order}")
        return dumpsite_index


    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # logger
        self.logger = mine.global_logger.get_logger("LLMDispatcher")
        # 获取当前卡车信息
        truck_load = truck.truck_load
        cur_location = truck.current_location.name
        cur_dumpsite = mine.get_dest_obj_by_name(cur_location)
        cur_dumpsite_index = mine.dump_sites.index(cur_dumpsite)
        assert isinstance(cur_dumpsite,
                          DumpSite), f"the truck {truck.name} is not in a dumpsite, it is in {cur_dumpsite.name}"

        # 统计dumpsite信息
        dump_sites = mine.dump_sites
        dumpsite_queue_length = [dump_site.parking_lot.queue_status["total"][int(mine.env.now)] for dump_site in
                                 dump_sites]
        estimated_dumpsite_queue_wait_times = [dump_site.estimated_queue_wait_time for dump_site in dump_sites]

        # 获取loadsite信息
        avaliable_loadsites = [loadsite for loadsite in mine.load_sites if loadsite.parking_lot is not None]
        load_site_names = [loadsite.name for loadsite in avaliable_loadsites]
        loadsite_queue_length = [loadsite.parking_lot.queue_status["total"][int(mine.env.now)] for loadsite in
                                 avaliable_loadsites]
        estimated_loadsite_queue_wait_times = [loadsite.estimated_queue_wait_time for loadsite in avaliable_loadsites]

        # 获取Road距离信息
        road_matrix = mine.road.road_matrix

        # 历史
        past_orders_back = [order for order in self.order_history if order["order_type"] == "back_order"][-10:]

        prompt = f"""
                    You are now an LLM scheduler, and your task is to assign the current truck a task to return to the loading area.
                    Background knowledge:
                    The mine has multiple loading and unloading areas as well as traffic roads. After being loaded with ore at a loading area, the mining truck needs to go to an unloading area to unload, and then return to the loading area to repeat the process.
                    Each unloading area has different unloading capabilities and queue situations, and each loading area also has different loading capabilities and queue situations.
                    The mining trucks are heterogeneous, with differences in their running speeds and loading capacities.
                    If a road has many mining trucks, the probability of random events like traffic jams and road maintenance increases, leading to longer operation times for the trucks.

                    Current mine information:
                    Unloading areas: {[{"name": dumpsite.name, "type": "dumpsite", "index": i, "queue_length": dumpsite_queue_length[i]
                                        } for i, dumpsite in enumerate(mine.dump_sites)]},
                    Loading areas: {[{"name": loadsite.name, "type": "loadsite", "index": j, "distance": road_matrix[j][cur_dumpsite_index],
                                      "queue_length": loadsite_queue_length[j]
                                      } for j, loadsite in enumerate(avaliable_loadsites)]},
                    Current road information:
                          {[{"road_id": f"{cur_location} to {loadsite.name}", "road_desc": f"from {cur_location} to {loadsite.name}", "distance": road_matrix[j][cur_dumpsite_index],
                             "trucks_on_this_road": mine.road.road_status[(cur_location, loadsite.name)]["truck_count"],
                             "jammed_trucks_on_this_road": mine.road.road_status[(cur_location, loadsite.name)]["truck_jam_count"],
                             "is_road_in_repair": mine.road.road_status[(cur_location, loadsite.name)]["repair_count"]}
                            for j, loadsite in enumerate(avaliable_loadsites)]}

                    Historical scheduling decisions:
                            {[{key: val for key, val in order.items() if key not in ['prompt', 'response']} for order in past_orders_back]}

                    The current truck is at the unloading area {cur_location} and needs to return to the loading area for loading.
                    Current truck order request information:
                    {{
                    "cur_time": {mine.env.now},
                    "order_type": "back_order",
                    "truck_location": "{truck.current_location.name}",
                    "truck_name": "{truck.name}",
                    "truck_capacity": {truck.truck_capacity},
                    "truck_speed": {truck.truck_speed}
                    }}

                    Please assign a suitable loading area as the target location for the current truck based on the information above, allowing it to return as quickly as possible for loading:
                    Requirements:
                        1. Overall objective: Considering the impact of random events on the road, choose the loading area with the shortest distance as much as possible, while avoiding traffic jams and road repairs.
                        2. Finally, based on the overall objective, directly provide the following JSON string as the decision result:
                    {{
                        "truck_name": "{truck.name}",
                        "loadsite_index": an integer number from 0 to {len(avaliable_loadsites) - 1}
                    }}
                """

        for i in range(3):
            try:
                response = self.OPENAI.get_response(prompt)
                self.logger.info(f"LLM 订单{self.order_index + 1}：prompt:{prompt} \n {response}")
                start = response.find('{')
                end = response.rfind('}') + 1
                json_str = response[start:end]
                data = json.loads(json_str)
                loadsite_index = data["loadsite_index"]
                break
            except Exception as e:
                print(e)
                loadsite_index = random.randint(0, len(avaliable_loadsites) - 1)
                self.logger.error(f"LLM 订单{self.order_index + 1}：parse error，giving random order")

        order = {
            "cur_time": mine.env.now,
            "order_type": "back_order",
            "truck_name": truck.name,
            "truck_capacity": truck.truck_capacity,
            "truck_speed": truck.truck_speed,
            "truck_location": f"{truck.current_location.name}",
            "loadsite_index": loadsite_index,
            "prompt": prompt,
            "response": response}
        self.back_order_history.append(order)
        self.order_history.append(order)
        self.order_index += 1
        self.logger.debug(f"LLM BACK 订单{self.order_index}：{order}")
        return loadsite_index


class OPENAI:
    def __init__(self, model_name="gpt-3.5-turbo"):
        self.api_key = "YOUR API KEY"  # token
        self.api_base = "OPENAI BASE HERE"  # you can choose custom api base, like:"https://api.qaqgpt.com/v1"
        self.model_name = model_name
        self.load_model()
    def load_model(self):
        openai.api_key = self.api_key
        openai.api_base = self.api_base

    def get_response(self, prompt):
        message =  [
                    {"role": "user", "content": prompt}]
        # time.sleep(1)
        response = openai.ChatCompletion.create(model=self.model_name, messages=message)
        for re in response["choices"]:
            return re["message"]["content"].strip()
        return ''


if __name__ == "__main__":
    dispatcher = LLMDispatcher()
    print(dispatcher.give_init_order(1,2))
    print(dispatcher.give_haul_order(1,2))
    print(dispatcher.give_back_order(1,2))

    print(dispatcher.total_order_count,dispatcher.init_order_count)


from __future__ import annotations
import numpy as np  # 导入NumPy库

from openmines.src.dispatcher import BaseDispatcher

class RandomDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "RandomDispatcher"
        self.np_random = np.random.RandomState()  # 创建NumPy的随机状态对象

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取可用的loadsite
        avaliable_loadsites = [loadsite for loadsite in mine.load_sites if loadsite.parking_lot is not None]
        # 随机一个数字
        random_index = self.np_random.randint(0, len(avaliable_loadsites))
        return random_index

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取可用的dumpsite
        avaliable_dumpsites = [loadsite for loadsite in mine.load_sites if loadsite.parking_lot is not None]
        random_index = self.np_random.randint(0, len(avaliable_dumpsites))
        return random_index

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取可用的loadsite
        avaliable_loadsites = [loadsite for loadsite in mine.dump_sites if loadsite.parking_lot is not None]
        # 随机一个数字
        random_index = self.np_random.randint(0, len(avaliable_loadsites))
        return random_index

if __name__ == "__main__":
    dispatcher = RandomDispatcher()
    print(dispatcher.give_init_order(1,2))
    print(dispatcher.give_haul_order(1,2))
    print(dispatcher.give_back_order(1,2))

    print(dispatcher.total_order_count,dispatcher.init_order_count)


import logging
import time, os
from pathlib import Path

# 项目主路径
PROJECT_ROOT_PATH = Path(__file__).parent.parent.parent
LOG_FILE_PATH = PROJECT_ROOT_PATH / "src" / "data" / "logs"

class MineLogger:
    def __init__(self, log_path=LOG_FILE_PATH, file_level=logging.DEBUG, console_level=logging.INFO):
        # 时间(兼容windows文件名)
        time_format = "%Y-%m-%d-%H-%M-%S"
        time_str = time.strftime(time_format, time.localtime())

        # LOGGER配置
        # 日志格式
        LOG_FORMAT = '%(asctime)s [%(levelname)s] %(name)s: %(message)s'

        # 日志级别
        CONSOLE_LOG_LEVEL = logging.INFO
        FILE_LOG_LEVEL = file_level

        # 控制台处理器
        self.CONSOLE_HANDLER = logging.StreamHandler()
        self.CONSOLE_HANDLER.setLevel(CONSOLE_LOG_LEVEL)
        self.CONSOLE_HANDLER.setFormatter(logging.Formatter(LOG_FORMAT))

        # 文件处理器
        self.log_path = log_path
        if not os.path.exists(self.log_path):
            os.makedirs(self.log_path)
        self.FILE_HANDLER = logging.FileHandler(self.log_path / f'openmines_sim_{time_str}.log')
        self.FILE_HANDLER.setLevel(FILE_LOG_LEVEL)
        self.FILE_HANDLER.setFormatter(logging.Formatter(LOG_FORMAT))

    def get_logger(self, name):
        """
        在初始化后，通过get_logger方法获取模块级logger
        :param name:
        :return:
        """
        logger = logging.getLogger(name)
        # 检查是否已经添加了控制台处理器
        if not any(isinstance(handler, logging.StreamHandler) for handler in logger.handlers):
            logger.addHandler(self.CONSOLE_HANDLER)
        # 检查是否已经添加了文件处理器
        if not any(isinstance(handler, logging.FileHandler) for handler in logger.handlers):
            logger.addHandler(self.FILE_HANDLER)
        logger.setLevel(logging.DEBUG)
        return logger


from __future__ import annotations

import time
from openmines.src.dispatcher import BaseDispatcher


class TabuSearchDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "TabuSearchDispatcher"

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        # 根据当前矿山状态，基于tabu search求解出订单
        ## 求解所需的矿山状态信息


        # 从第一个load site开始
        return 0

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        # 从第一个load site开始
        return self.solution

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # 从第一个dump site开始
        return 0

if __name__ == "__main__":
    dispatcher = TabuSearchDispatcher()
    print(dispatcher.give_init_order(1,2))
    print(dispatcher.give_init_order(1, 2))
    print(dispatcher.give_init_order(1, 2))
    print(dispatcher.give_init_order(1, 2))
    print(dispatcher.give_haul_order(1,2))
    print(dispatcher.give_back_order(1,2))

    print(dispatcher.total_order_count,dispatcher.init_order_count,dispatcher.init_order_time,dispatcher.total_order_time)


import random

import numpy as np
import simpy
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
from simpy.resources.resource import Request



from openmines.src.charging_site import ChargingSite
from openmines.src.dump_site import DumpSite, Dumper
from openmines.src.load_site import LoadSite, Shovel
from openmines.src.utils.event import Event, EventPool

TRUCK_DEFAULT_SPEED = 25 # km/h
"""在仿真结束后将仿真过程中矿卡的位置数据按帧转换成DataFrame并保存到文件中
        卡车地理标号(start,dest)：
            -1 代表充电区
            0-N 代表装载区
            N+1 N+M 代表卸载区
        卡车状态(state)：
            0 代表空载
            1 代表等载
            2 代表装载
            3 代表满载
            4 代表等卸
            5 代表卸载
        卡车位置数值(position_ratio)：
            0-1 代表从当前点到目标点的比率（仅在0 3过程有用）
        卡车位置坐标(position_x, position_y)：
            装载点 卸载点 铲车等等的位置 我们认为其已经在配置文件中提供了
            这里是计算好的二维坐标位置
            TODO：另外我们可视化的时候需要考虑到多个车辆重叠的情况(引入仿真sumo？)
        卡车时间(time)：
            仿真时间
        """

class LoadRequest(Request):
    """
    自定义的simpy资源请求类，用于在卡车到达装载区时请求铲车资源；
    """
    def __init__(self, resource, truck, load_site):
        super().__init__(resource)
        self.truck = truck
        self.load_site = load_site

class DumpRequest(Request):
    def __init__(self, resource, truck, dump_site):
        super().__init__(resource)
        self.truck = truck
        self.dump_site = dump_site

class Truck:
    def __init__(self, name:str, truck_capacity:float, truck_speed:float=TRUCK_DEFAULT_SPEED):
        self.name = name
        self.truck_capacity = truck_capacity  # truck capacity in tons
        self.truck_speed = truck_speed  # truck speed in km/h
        self.current_location = None
        self.target_location = None
        self.journey_start_time = 0
        self.journey_coverage = None
        self.last_breakdown_time = 0
        self.event_pool = EventPool()
        # the probalistics of the truck
        self.expected_working_time_without_breakdown = 60*6  # in minutes
        self.expected_working_time_until_unrepairable = 60*24*7*7  # in minutes
        self.repair_avg_time = 10
        self.repair_std_time = 3
        # truck status
        self.status = "idle"
        self.truck_load = 0  # in tons, the current load of the truck
        self.service_count = 0  # the number of times the truck has been dumped
        self.total_load_count = 0  # the total load count of the truck
        self.truck_cycle_time = 0
        self.first_order_time = 0
        # RL
        self.current_decision_event = None

    def set_env(self, mine:"Mine"):
        self.mine = mine
        self.env = mine.env
        self.dispatcher = mine.dispatcher

    def move(self, target_location, distance:float, manual_speed:float=None):
        """
        移动函数 需要区分满载和空载
        :param target_location:
        :param distance:
        :param manual_speed:
        :return:
        """
        assert target_location is not None, "target_location can not be None"
        assert target_location is not self.current_location, "target_location can not be the same as current_location"
        assert distance >= 0, "distance can not be negative"
        assert self.truck_speed > 0, "truck_speed can not be negative"
        # 记录当前路程的信息
        self.target_location = target_location
        self.journey_start_time = self.env.now
        if manual_speed is not None:
            assert manual_speed > 0, "manual_speed can not be negative"
            duration = (distance / manual_speed)*60  # in minutes
        else:
            duration = (distance / self.truck_speed)*60
        self.truck_speed = manual_speed if manual_speed is not None else self.truck_speed

        """
        1.车辆维修随机事件的模拟        
        """
        # 检查车辆是否发生故障并获得维修时间
        repair_time = self.check_vehicle_availability()
        if repair_time:
            # 如果发生故障，记录故障事件并进行维修
            breakdown_event = Event(self.last_breakdown_time, "TruckEvent:breakdown", f'Time:<{self.last_breakdown_time}> Truck:[{self.name}] breakdown for {repair_time} minutes',
                      info={"name": self.name, "status": "breakdown",
                            "start_time": self.env.now, "end_time": self.env.now + repair_time})
            self.event_pool.add_event(breakdown_event)
            self.mine.random_event_pool.add_event(breakdown_event)
            self.logger.info(f'Time:<{self.last_breakdown_time}> Truck:[{self.name}] breakdown for {repair_time} minutes at {self.last_breakdown_time}')
            # 进行维修（暂停运行）
            self.status = "repairing"
            yield self.env.timeout(repair_time)
        """
        2.车辆运行过程中的堵车随机事件模拟
        """
        # 分析当前道路中的车辆情况，并随机生成一个延迟时间用来模拟交通堵塞
        # 获取其他车辆的引用
        other_trucks = self.mine.trucks
        # 筛选出出发地和目的地跟本车相同的车辆
        other_trucks_on_road = [truck for truck in other_trucks if truck.current_location == self.current_location and
                                truck.target_location == self.target_location]
        # 为每个同路的车辆根据当前时间计算route_coverage数值
        for truck in other_trucks_on_road:
            truck.journey_coverage = truck.get_route_coverage(distance)
        # 假设每个车辆导致堵车的概率在route_coverage上满足正态分布，均值为route_coverage，方差为0.1
        # other_trucks_on_road中的每个车导致堵车的概率会叠加在一起取平均成为最终的堵车概率分布
        truck_positions = [truck.journey_coverage for truck in other_trucks_on_road]
        # 每个卡车的正态分布的标准差
        sd = 0.1  # 根据3sigma原则，这个值代表这个车辆对堵车的影响会在+-0.3总路程的范围内
        # 位置的范围
        x = np.linspace(0, 1, 1000)
        # 初始化堵车的概率为0
        total_prob = np.zeros_like(x)
        # 对每辆卡车
        for i, position in enumerate(truck_positions):
            # 计算在每个位置的堵车概率
            prob = norm.pdf(x, position, sd)
            # 将这个卡车的堵车概率加到总的堵车概率上
            total_prob = total_prob + prob
        # 正规化总的堵车概率
        total_prob = total_prob / len(truck_positions)
        # 使用蒙特卡洛方法来决定是否会发生堵车
        try:
            if np.sum(total_prob) == 0:
                probabilities = np.ones_like(total_prob) / len(total_prob)
            else:
                probabilities = total_prob / (np.sum(total_prob))
            jam_position = np.random.choice(x, p=probabilities)
        except:
            print("bug")
        # 在jam_position处使用weibull分布，采样一个可能的堵车时间
        jam_time = np.random.weibull(2) * 10
        time_to_jam = jam_position * duration
        if time_to_jam < jam_time:
            # 如果到达堵车区域的时间小于堵车时间，那么就会发生堵车
            is_jam = True if len(truck_positions) > 3 else False
            self.mine.random_event_pool.add_event(Event(self.env.now + time_to_jam, "RoadEvent:jam",
                                                        f'Time:<{self.env.now + time_to_jam}> Truck:[{self.name}] is jammed at road from {self.current_location.name} '
                                                        f'to {self.target_location.name} for {jam_time:.2f} minutes',
                                                        info={"name": self.name, "status": "jam", "speed": 0,
                                                              "start_location": self.current_location.name,
                                                                "end_location": self.target_location.name,
                                                              "start_time": self.env.now, "est_end_time": self.env.now + time_to_jam + jam_time}))
            self.logger.info(f"Time:<{self.last_breakdown_time}> Truck:[{self.name}] is jammed at {self.current_location.name} to {self.target_location.name} for {jam_time:.2f} minutes")
        else:
            is_jam = False


        """
        3.车辆完全损坏的模拟(当车辆完全损坏就会被移除,车辆到指定时间后不会再发生移动，而是停留在原地
        """
        if repair_time is None:
            self.logger.info(f"Time:<{self.last_breakdown_time}> Truck:[{self.name}] is broken down and beyond repair at {self.current_location.name} to "
                             f"{self.target_location.name}")
            unrepairable_event = Event(self.env.now, "TruckEvent:unrepairable", f'Time:<{self.env.now}> Truck:[{self.name}] is broken down and beyond repair'
                                                                     f' at {self.current_location.name} to '
                                                                        f'{self.target_location.name}',
                                        info={"name": self.name, "status": "unrepairable","time": self.env.now})
            self.event_pool.add_event(unrepairable_event)
            self.mine.random_event_pool.add_event(unrepairable_event)
            self.status = "unrepairable"
            return
        """
        4.车辆移动前往目的地的模拟
        """
        # 判断目标地点的类型
        if isinstance(self.current_location, DumpSite) and isinstance(target_location, LoadSite):
            event_name = "unhaul"
        elif isinstance(self.current_location, ChargingSite):
            event_name = "init"
        else:
            event_name = "haul"

        self.event_pool.add_event(Event(self.env.now, event_name, f'Truck:[{self.name}] moves at {target_location.name}',
                                        info={"name": self.name, "status": event_name, "speed": manual_speed if manual_speed is not None else self.truck_speed,
                                              "start_time": self.env.now, "est_end_time": self.env.now+duration, "end_time": None,
                                              "start_location": self.current_location.name,
                                              "target_location": target_location.name,
                                              "distance": distance, "duration": None}))
        self.status = "moving"
        yield self.env.timeout(duration+is_jam*jam_time)
        # 补全数据
        last_move_event = self.event_pool.get_last_event(type=event_name, strict=True)
        last_move_event.info["end_time"] = self.env.now
        last_move_event.info["duration"] = self.env.now - last_move_event.info["start_time"]
        # after arrival set current location
        assert type(self.current_location) != type(target_location), f"current_site and target_site should not be the same type of site "
        self.current_location = target_location

    def load(self, shovel:Shovel):
        shovel_tons = shovel.shovel_tons
        shovel_cycle_time = shovel.shovel_cycle_time
        load_time = (self.truck_capacity/shovel_tons) * shovel_cycle_time
        self.logger.info(f'Time:<{self.env.now}> Truck:[{self.name}] Start loading at shovel {shovel.name}, load time is {load_time}')
        self.status = "loading"
        yield self.env.timeout(load_time)
        # 随机生成一个装载量 +-10%
        self.truck_load = self.truck_capacity*(1+np.random.uniform(-0.1, 0.1))
        shovel.produced_tons += self.truck_load
        shovel.service_count += 1
        self.logger.info(f'Time:<{self.env.now}> Truck:[{self.name}] Finish loading at shovel {shovel.name}')

    def unload(self, dumper:Dumper):
        unload_time:float = dumper.dump_time
        self.status = "unloading"
        yield self.env.timeout(unload_time)
        dumper.dumper_tons += self.truck_load
        self.total_load_count += self.truck_load
        self.service_count += 1
        self.truck_cycle_time = (self.env.now - self.last_breakdown_time) / self.service_count
        self.truck_load = 0
        dumper.service_count += 1
        self.logger.info(f'Time:<{self.env.now}> Truck:[{self.name}] Finish unloading at dumper {dumper.name}, dumper tons is {dumper.dumper_tons}')
        # self.event_pool.add_event(Event(self.env.now, "unload", f'Truck:[{self.name}] Finish unloading at dumper {dumper.name}',
        #                           info={"name": self.name, "status": "unloading",
        #                                 "start_time": self.env.now-unload_time, "end_time": self.env.now,
        #                                 "dumper": dumper.name, "unload_duration": unload_time}))
        # WARN：这里unload和get dumper重复了

    def check_queue_position(self, shovel, request):
        """
        检查并返回矿车在等待队列中的位置
        :param shovel: 要检查队列的铲车
        :param request: 矿车的资源请求对象
        :return: 队列中的位置（从0开始计数）
        """
        try:
            if len(shovel.res.queue) == 0:
                return 0
            elif shovel.res.queue.index(request) == 0:
                return 0
            else:
                return shovel.res.queue.index(request)
        except ValueError:
            return 0  # 如果请求不在队列中，则返回-1

    def wait_for_decision(self):
        # 创建一个等待事件
        self.current_decision_event = self.env.event()
        yield self.current_decision_event  # 等待rl环境决策
        return self.current_decision_event.value  # 返回决策结果

    def run(self, is_rl_training=False):
        """
        矿车运行的主干入口函数
        车辆从充电区开始，然后请求init_order前往装载区
        到达装载区后，装载区分配一个shovel给车辆，然后车辆开始装载
        装载完成后，车辆请求haul_order前往卸载区
        到达卸载区后，卸载区分配一个dumper给车辆，然后车辆开始卸载
        卸载完成后，车辆请求back_order前往装载区
        TODO：1.到达时间的正态建模 2.电车油车的区分，油耗电耗统计 3.其他随机事件 4.统计图表
        :return:
        """
        # 配置日志
        self.logger = self.mine.global_logger.get_logger("Truck")

        # 轮班开始 车辆从充电区域前往装载区
        self.current_location = self.mine.charging_site
        self.status = "waiting for init order"
        if is_rl_training:
            # 等待RL agent给出决策;
            dest_load_index: int = yield self.env.process(self.wait_for_decision())
        else:
            dest_load_index: int = self.dispatcher.give_init_order(truck=self, mine=self.mine)  # TODO:允许速度规划
        self.first_order_time = self.env.now
        self.target_location = self.mine.load_sites[dest_load_index]

        move_distance:float = self.mine.road.charging_to_load[dest_load_index]
        load_site: LoadSite = self.mine.load_sites[dest_load_index]
        self.logger.info(f'Time:<{self.env.now}> Truck:[{self.name}] Activated at {self.env.now}, Target load site is ORDER({dest_load_index}):{load_site.name}, move distance is {move_distance}')
        self.event_pool.add_event(Event(self.env.now, "INIT ORDER", f'Truck:[{self.name}] Activated at {self.env.now}, Target load site is ORDER({dest_load_index}):{load_site.name}, move distance is {move_distance}',
                                        info={"name": self.name, "status": "INIT ORDER",
                                              "start_time": self.env.now, "est_end_time": self.env.now+move_distance/self.truck_speed,
                                              "start_location": self.current_location.name,
                                              "target_location": load_site.name,
                                              "distance": move_distance, "order_index": dest_load_index}))
        yield self.env.process(self.move(target_location=load_site, distance=move_distance))  # 移动时间
        # 检查车辆状态，如果是完全损坏则退出
        if self.status == "unrepairable":
            self.logger.info(f"Truck {self.name} is beyond repair and will no longer participate in operations.")
            return

        while True:
            # 到达装载区开始请求资源并装载
            self.logger.info(f'Time:<{self.env.now}> Truck:[{self.name}] Arrived at {self.mine.load_sites[dest_load_index].name} at {self.env.now}')
            load_site:LoadSite = self.mine.load_sites[dest_load_index]
            shovel = load_site.get_available_shovel()

            with LoadRequest(shovel.res, self, load_site) as req:
                # 申请到资源之前的操作
                truck_queue_index = self.check_queue_position(shovel, req)
                self.event_pool.add_event(Event(self.env.now, "wait shovel", f'Truck:[{self.name}] Wait shovel {shovel.name}',
                                                info={"name": self.name, "status": "waiting for shovel",
                                                      "queue_index": truck_queue_index,
                                                      "start_time": self.env.now, "end_time": None,
                                                      "shovel": shovel.name, "wait_duration": None}))
                self.status = "waiting for shovel"
                # ...
                yield req  # 申请铲车资源
                # 申请到铲车资源
                # 先获取wait的event进行数据补全
                last_wait_event = self.event_pool.get_last_event(type="wait shovel", strict=True)
                last_wait_event.info["end_time"] = self.env.now
                last_wait_event.info["wait_duration"] = self.env.now - last_wait_event.info["start_time"]
                # 添加load的event
                self.event_pool.add_event(Event(self.env.now, "get shovel", f'Truck:[{self.name}] Get shovel {shovel.name}',
                                                info={"name": self.name, "status": "loading on shovel",
                                                      "start_time": self.env.now, "end_time": None,
                                                      "shovel": shovel.name, "load_duration": None}))
                yield self.env.process(self.load(shovel))  # 装载时间 同shovel和truck自身有关系
                # 装载完毕 对之前的Event进行数据补全
                last_load_event = self.event_pool.get_last_event(type="get shovel", strict=True)
                last_load_event.info["end_time"] = self.env.now
                last_load_event.info["load_duration"] = self.env.now - last_load_event.info["start_time"]

            # 装载完毕，请求新的卸载区，并开始移动到卸载区
            self.status = "waiting for haul order"
            if is_rl_training:
                # 等待RL agent给出决策;
                dest_unload_index: int = yield self.env.process(self.wait_for_decision())
            else:
                dest_unload_index: int = self.dispatcher.give_haul_order(truck=self, mine=self.mine)
            dest_unload_site: DumpSite = self.mine.dump_sites[dest_unload_index]
            move_distance: float = self.mine.road.get_distance(truck=self, target_site=dest_unload_site)
            self.target_location = dest_unload_site

            self.logger.debug(f"Time:<{self.env.now}> Truck:[{self.name}] Start moving to ORDER({dest_unload_index}): {dest_unload_site.name}, move distance is {move_distance}, speed: {self.truck_speed}")
            self.event_pool.add_event(Event(self.env.now, "ORDER", f'Truck:[{self.name}] Start moving to ORDER({dest_unload_index}): {dest_unload_site.name}, move distance is {move_distance}, speed: {self.truck_speed}',
                                            info={"name": self.name, "status": "ORDER",
                                                  "start_time": self.env.now, "est_end_time": self.env.now+move_distance/self.truck_speed,
                                                  "start_location": self.current_location.name,
                                                  "target_location": dest_unload_site.name, "speed": self.truck_speed,
                                                  "distance": move_distance, "order_index": dest_unload_index}))

            yield self.env.process(self.move(target_location=dest_unload_site, distance=move_distance))  # 移动时间

            # 检查车辆状态，如果是完全损坏则退出
            if self.status == "unrepairable":
                self.logger.info(f"Truck {self.name} is beyond repair and will no longer participate in operations.")
                return
            # 到达卸载区并开始请求资源并卸载
            self.logger.debug(f'Time:<{self.env.now}> Truck:[{self.name}] Arrived at {dest_unload_site.name} at {self.env.now}')
            dumper:Dumper = dest_unload_site.get_available_dumper()
            with DumpRequest(dumper.res, self, dest_unload_site) as req:
                # 申请到资源之前的操作
                # ...
                self.event_pool.add_event(Event(self.env.now, "wait dumper", f'Truck:[{self.name}] Wait dumper {dumper.name}',
                                                info={"name": self.name, "status": "waiting for dumper",
                                                      "start_time": self.env.now, "end_time": None,
                                                      "dumper": dumper.name, "wait_duration": None}))
                self.status = "waiting for dumper"
                yield req  # 申请卸载位资源
                # 申请到卸载位资源
                # 先获取wait的event进行数据补全
                last_wait_event = self.event_pool.get_last_event(type="wait dumper", strict=True)
                last_wait_event.info["end_time"] = self.env.now
                last_wait_event.info["wait_duration"] = self.env.now - last_wait_event.info["start_time"]
                self.event_pool.add_event(Event(self.env.now, "get dumper", f'Truck:[{self.name}] Get dumper {dumper.name}',
                                                info={"name": self.name, "status": "unloading on dumper",
                                                      "start_time": self.env.now, "end_time": None,
                                                      "dumper": dumper.name, "unload_duration": None}))
                yield self.env.process(self.unload(dumper))
                # 卸载完毕 对之前的Event进行数据补全
                last_unload_event = self.event_pool.get_last_event(type="get dumper", strict=True)
                last_unload_event.info["end_time"] = self.env.now
                last_unload_event.info["unload_duration"] = self.env.now - last_unload_event.info["start_time"]

            # 卸载完毕，请求新的装载区，并开始移动到装载区
            self.status = "waiting for back order"
            if is_rl_training:
                # 等待RL agent给出决策;
                dest_load_index: int = yield self.env.process(self.wait_for_decision())
            else:
                dest_load_index: int = self.dispatcher.give_back_order(truck=self, mine=self.mine)
            dest_load_site: LoadSite = self.mine.load_sites[dest_load_index]
            move_distance: float = self.mine.road.get_distance(truck=self, target_site=dest_load_site)
            self.target_location = dest_load_site
            self.logger.debug(f"Time:<{self.env.now}> Truck:[{self.name}] Start moving to ORDER({dest_load_index}):{dest_load_site.name}, move distance is {move_distance}, speed: {self.truck_speed}")
            self.event_pool.add_event(Event(self.env.now, "ORDER", f'Truck:[{self.name}] Start moving to ORDER({dest_load_index}):{dest_load_site.name}, move distance is {move_distance}, speed: {self.truck_speed}',
                                            info={"name": self.name, "status": "ORDER",
                                                  "start_time": self.env.now, "est_end_time": self.env.now+move_distance/self.truck_speed,
                                                  "start_location": self.current_location.name,
                                                  "target_location": dest_load_site.name, "speed": self.truck_speed,
                                                  "distance": move_distance, "order_index": dest_load_index}))
            yield self.env.process(self.move(target_location=dest_load_site, distance=move_distance))  # 移动时间
            # 检查车辆状态，如果是完全损坏则退出
            if self.status == "unrepairable":
                self.logger.info(f"Truck {self.name} is beyond repair and will no longer participate in operations.")
                return

    def charge(self, duration):
        """
        电矿卡需要充电
        油矿卡不需要充电
        TODO：针对卡车油耗、电耗、充电时间进行建模。并应用。
        :param duration:
        :return:
        """
        self.logger.info(f'{self.name} Start charging at {self.env.now}')
        yield self.env.timeout(duration)

    def get_wait_time(self):
        """
        获取一次循环的等待时间
        目前是计算unload事件的个数和时间，然后计算平均值
        :return:
        """
        # TODO:不能只有接到资源的车才计算wait时间 所有的车都应该计算wait时间 有的根本没有获得资源 15811296389
        wait_shovel_events = self.event_pool.get_even_by_type("wait shovel")
        end_wait_shovel_events = self.event_pool.get_even_by_type("get shovel")
        self.event_pool.add_event(Event(self.mine.total_time, "end", f'Truck:[{self.name}] End'))
        end_event = self.event_pool.get_even_by_type("end")

        wait_shovel_event_count = len(wait_shovel_events)
        end_wait_shovel_event_count = len(end_wait_shovel_events)

        if wait_shovel_event_count > end_wait_shovel_event_count:
            # 如果等待铲车的事件比获得铲车的事件多，说明结束时间打断了等待铲车的事件，为结束等待添加一个结束时间
            end_wait_shovel_events.append(end_event[0]) # end evnet 没有添加成功

        wait_shovel_event_time = sum([event.time_stamp for event in wait_shovel_events])
        end_wait_shovel_event_time = sum([event.time_stamp for event in end_wait_shovel_events])
        wait_time = (end_wait_shovel_event_time - wait_shovel_event_time) / wait_shovel_event_count
        if wait_time == float('inf'):
            wait_time = 0
        return wait_time

    def get_route_coverage(self,distance)->float:
        """
        获取一次循环的路程覆盖率 0-1
        0代表刚刚出发 1代表完成
        这里只是一种近似，没有考虑到当前路途的交通情况
        :return:
        """
        assert distance > 0, "distance must be greater than 0"
        if self.journey_start_time is None:
            print(1)
        assert self.journey_start_time is not None, "journey_start_time must be not None, is the truck journey started?"

        # 获取当前环境时间
        current_time = self.env.now
        # 获取卡车的速度
        speed = self.truck_speed
        # 获取卡车的总共预期行驶时间
        total_travel_time = (distance/speed)*60  # 单位：分钟
        # 获取卡车的路程覆盖率
        coverage = (current_time - self.journey_start_time)/total_travel_time
        return coverage

    def check_vehicle_availability(self):
        """
        使用指数分布对卡车的可用性进行建模并采样，可能出现的状况包括故障需要维修。
        :return: 维修持续时间（如果发生故障)
        """
        # 使用指数分布计算故障发生的时间（载具正常工作时长）
        time_to_breakdown = np.random.exponential(self.expected_working_time_without_breakdown)
        # 判断载具是否发生故障（当前时间超过故障时间）
        if self.env.now >= self.last_breakdown_time + time_to_breakdown:
            # 发生故障，使用正态分布N(mu=10, sigma=3)采样获得维修时间
            repair_time = np.random.normal(self.repair_avg_time, self.repair_std_time)
            repair_time = max(repair_time, 0)  # 确保维修时间为正值
            # 更新最后一次故障时间
            self.last_breakdown_time = self.last_breakdown_time + time_to_breakdown
            """
            TODO: 维修的判断依据有问题 很有可能多次采样叠加后self.last_breakdown_time依然是小于当前时间的
            这里明天再修复 今天弄不了 脑子不好用了
            """
            return repair_time  # 发生故障则返回维修时间
        # 如果车辆发生无法维修的损坏，返回None；使用指数分布计算故障发生的时间
        # self.expected_working_time_until_unrepairable

        if self.env.now >= random.expovariate(1.0 / self.expected_working_time_until_unrepairable):
            return None
        return 0  # 未发生故障则返回0维修时间



import argparse
import json
import pathlib
import sys
import pkgutil
import importlib
from datetime import time

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from openmines.src.utils.visualization.charter import Charter
from openmines.src.utils.visualization.graphher import VisualGrapher

# add the openmines package to the path
sys.path.append(str(pathlib.Path(__file__).parent.parent.parent.parent.absolute()))

from openmines.src.mine import Mine
from openmines.src.truck import Truck
from openmines.src.charging_site import ChargingSite
from openmines.src.load_site import LoadSite, Shovel
from openmines.src.dump_site import DumpSite, Dumper
from openmines.src.road import Road
from openmines.src.dispatcher import BaseDispatcher
from openmines.src.dispatch_algorithms.naive_dispatch import NaiveDispatcher
from openmines.src.dispatch_algorithms import *

def load_config(filename):
    with open(filename, 'r') as file:
        return json.load(file)

def run_dispatch_sim(dispatcher: BaseDispatcher, config_file):
    config = load_config(config_file)
    # log_path 为cwd下的logs文件夹
    log_path = pathlib.Path.cwd() / 'logs'
    # 初始化矿山
    mine = Mine(config['mine']['name'], log_path=log_path)
    mine.add_dispatcher(dispatcher)
    # 初始化充电站和卡车
    charging_site = ChargingSite(config['charging_site']['name'], position=config['charging_site']['position'])
    for truck_config in config['charging_site']['trucks']:
        for _ in range(truck_config['count']):
            truck = Truck(
                name=f"{truck_config['type']}{_ + 1}",
                truck_capacity=truck_config['capacity'],
                truck_speed=truck_config['speed']
            )
            charging_site.add_truck(truck)

    # 初始化装载点和铲车
    for load_site_config in config['load_sites']:
        load_site = LoadSite(name=load_site_config['name'], position=load_site_config['position'])
        for shovel_config in load_site_config['shovels']:
            shovel = Shovel(
                name=shovel_config['name'],
                shovel_tons=shovel_config['tons'],
                shovel_cycle_time=shovel_config['cycle_time'],
                position_offset=shovel_config['position_offset']
            )
            load_site.add_shovel(shovel)
        load_site.add_parkinglot(position_offset=load_site_config['parkinglot']['position_offset'],
                                 name=load_site_config['parkinglot']['name'])
        mine.add_load_site(load_site)

    # 初始化卸载点和卸载机
    for dump_site_config in config['dump_sites']:
        dump_site = DumpSite(dump_site_config['name'], position=dump_site_config['position'])
        for dumper_config in dump_site_config['dumpers']:
            for _ in range(dumper_config['count']):
                dumper = Dumper(
                    name=f"{dump_site_config['name']}-点位{_}",
                    dumper_cycle_time=dumper_config['cycle_time'],
                    position_offset=dumper_config['position_offset']
                )
                dump_site.add_dumper(dumper)
        dump_site.add_parkinglot(position_offset=dump_site_config['parkinglot']['position_offset'],
                                 name=dump_site_config['parkinglot']['name'])
        mine.add_dump_site(dump_site)

    # 初始化道路
    road_matrix = np.array(config['road']['road_matrix'])
    road_event_params = config['road'].get('road_event_params', {})  # 从配置中加载道路事件参数

    charging_to_load_road_matrix = config['road']['charging_to_load_road_matrix']
    road = Road(road_matrix=road_matrix, charging_to_load_road_matrix=charging_to_load_road_matrix,
                road_event_params=road_event_params)
    # # 添加充电站和装载区卸载区
    mine.add_road(road)
    mine.add_charging_site(charging_site)

    # 开始实验
    print(f"Running simulation for {dispatcher.__class__.__name__}")
    ticks = mine.start(total_time=config['sim_time'])
    return ticks


def run_simulation(config_file=None):
    config = load_config(config_file)
    charter = Charter(config_file)
    states_dict = dict()
    # 初始化调度器
    # 定义dispatch_algorithms所在的路径
    dispatchers_package = 'openmines.src.dispatch_algorithms'
    # 导入dispatch_algorithms包
    dispatchers_module = importlib.import_module(dispatchers_package)
    # 遍历dispatch_algorithms包中的所有模块
    dispatchers_list = []
    for _, module_name, _ in pkgutil.iter_modules(dispatchers_module.__path__, dispatchers_package + '.'):
        # 动态导入模块
        module = importlib.import_module(module_name)
        # 假设配置中指定的调度器类型
        dispatcher_types:list = config['dispatcher']['type']
        for dispatcher_type in dispatcher_types:
            # 检查模块中是否有该类
            if hasattr(module, dispatcher_type):
                dispatcher_class = getattr(module, dispatcher_type)
                dispatcher = dispatcher_class()
                if dispatcher.name not in [dispatcher.name for dispatcher in dispatchers_list]:
                    dispatchers_list.append(dispatcher)

    # 开始运行对比实验
    for dispatcher in dispatchers_list:
        dispatcher_name = dispatcher.name
        # RUN SIMULATION
        ticks = run_dispatch_sim(dispatcher, config_file)
        # 读取运行结果并保存，等待绘图
        ## 读取production
        times = []
        produced_tons_list = []
        service_count_list = []
        waiting_truck_count_list = []
        # ticks 是一个字典 key为时间，value为一个字典，包含了当前时间的所有信息
        for tick in ticks.values():
            if 'mine_states' not in tick:
                continue
            tick = tick['mine_states']
            times.append(tick['time'])
            produced_tons_list.append(tick['produced_tons'])
            service_count_list.append(tick['service_count'])
            waiting_truck_count_list.append(tick['waiting_truck_count'])
        states_dict[dispatcher_name] = {
            'times': times,
            'produced_tons_list': produced_tons_list,
            'service_count_list': service_count_list,
            'waiting_truck_count_list': waiting_truck_count_list,
            'summary': ticks['summary']
        }
    # 绘制图表
    charter.draw(states_dict)
    charter.save()

def run_visualization(tick_file=None):
    visual_grapher = VisualGrapher(tick_file)
    # 构造路径和文件名字
    gif_file = tick_file.strip('.json') + '.gif'
    visual_grapher.create_animation(output_path=gif_file)

def main():
    parser = argparse.ArgumentParser(description='Run a dispatch simulation of a mine with your DISPATCH algorithm and MINE config file')
    subparsers = parser.add_subparsers(help='commands', dest='command')

    # 直接访问入口 sisymine -f config.json, sisymine -v tick.json
    # 添加运行参数 '-f' 和 '-v'
    parser.add_argument('-f', '--config-file', type=str, help='Path to the config file')
    parser.add_argument('-v', '--tick-file', type=str, help='Path to the simulation tick file')

    # add command 'run'
    run_parser = subparsers.add_parser('run', help='Run a simulation experiment')
    run_parser.add_argument('-f', '--config-file', type=str, required=True, help='Path to the config file')

    # add command visualize
    visualize_parser = subparsers.add_parser('visualize', help='Visualize a simulation experiment')
    visualize_parser.add_argument('-f', '--tick-file', type=str, required=True, help='Path to the simulation tick file')

    args = parser.parse_args()
    # 如command为空，那么检查f/v参数是否存在，如果不存在则print help；如果存在f/v参数则执行run/visualize
    if args.command is None:
        if args.config_file is None and args.tick_file is None:
            parser.print_help()
        elif args.config_file is not None:
            run_simulation(config_file=args.config_file)
        elif args.tick_file is not None:
            run_visualization(tick_file=args.tick_file)
    if args.command == 'run':
        print("args.config_file", args.config_file)
        run_simulation(config_file=args.config_file)
    if args.command == 'visualize':
        tick_file = args.tick_file
        run_visualization(tick_file=tick_file)


if __name__ == "__main__":
    config_path = sys.argv[1]
    run_simulation(config_file=config_path)






class Event:
    def __init__(self, time_stamp:float, event_type:str, desc:str, info:dict=None):
        self.time_stamp = time_stamp
        self.event_type = event_type
        self.desc = desc
        self.info = info

    def __str__(self):
        return f"Event(time_stamp={self.time_stamp},event_type={self.event_type},desc={self.desc},info={self.info})"

    def __repr__(self):
        return self.__str__()


class EventPool:
    """
    用来存储卡车运行事件的类，便于后期统计分析
    """
    def __init__(self):
        self.event_set = dict()

    def add_event(self, event:Event):
        if event.time_stamp in self.event_set.keys():
            event.time_stamp = event.time_stamp + 0.0001
        self.event_set[event.time_stamp] = event

    def get_even_by_type(self, name:str, )->list:
        """
        通过事件type获取事件
        :param name:
        :return:
        """
        list_event = []
        for t in sorted(self.event_set.keys()):
            if name in self.event_set[t].event_type:
                list_event.append(self.event_set[t])
        return list_event

    def get_even_by_desc(self, name:str)->list:
        """
        通过事件desc获取事件
        :param name:
        :return:
        """
        list_event = []
        for t in sorted(self.event_set.keys()):
            if name in self.event_set[t].desc:
                list_event.append(self.event_set[t])
        return list_event

    def get_event_by_time(self,time:float,mode="backward")->list:
        """
        给定一个时间，获取当前时间之前的顺序的event列表
        :param time:
        :return:
        """
        if mode == "backward":
            list_event = []
            for t in sorted(self.event_set.keys()):
                if t <= time:
                    list_event.append(self.event_set[t])
            return list_event
        else:
            list_event = []
            for t in sorted(self.event_set.keys()):
                if t > time:
                    list_event.append(self.event_set[t])
            return list_event

    def get_event_by_time_range(self, start_time:float, end_time:float)->list:
        """
        给定一个时间范围，获取当前时间之前的顺序的event列表
        :param time:
        :return:
        """
        list_event = []
        for t in sorted(self.event_set.keys()):
            if start_time <= t <= end_time:
                list_event.append(self.event_set[t])
        return list_event

    def update_last_info(self, type:str, info:dict,strict:bool=True):
        """
        更新最后一个事件的info
        如果strict为True，则
            断言这个事件的type是给定type
        如果strict为False，则
            更新最近的匹配到的event的info

        :param type:
        :param info:
        :param strict:
        :return:
        """
        if strict:
            assert self.event_set[list(self.event_set.keys())[-1]].event_type == type
            self.event_set[list(self.event_set.keys())[-1]].info = info
        else:
            for t in sorted(self.event_set.keys(),reverse=True):
                if type in self.event_set[t].event_type:
                    self.event_set[t].info = info
                    break

    def get_last_event(self,type:str,strict:bool=True):
        """
        获取最后一个事件
        如果strict为True，则
            断言这个事件的type是给定type
        如果strict为False，则
            更新最近的匹配到的event的info
        """
        if strict:
            assert self.event_set[list(self.event_set.keys())[-1]].event_type == type
            return self.event_set[list(self.event_set.keys())[-1]]
        else:
            for t in sorted(self.event_set.keys(),reverse=True):
                if type in self.event_set[t].event_type:
                    return self.event_set[t]
                    break

    def clear(self):
        self.event_set.clear()

class RandomEventPool:
    """
    1.用来记录在模拟运行过程中的随机事件以供后期分析
    2.在算法派车决策时，提供现有的随机事件
    """
    pass


if __name__ == "__main__":
    # test EventPool
    print("==============test EventPool================")
    pool = EventPool()
    pool.add_event(Event(3.0, 'test', 'desc'))
    pool.add_event(Event(1.0, 'test', 'desc'))
    pool.add_event(Event(1.0, 'test2', 'desc'))
    pool.add_event(Event(2.0, 'dasdasd', 'desc'))

    print(pool.get_even_by_type('test'))
    print(pool.get_even_by_desc('desc'))
    # clear
    pool.clear()
    # test update_last_info
    print("==============test update_last_info================")
    pool.add_event(Event(1.0, 'move', 'desc'))
    pool.add_event(Event(1.5, 'wait', 'desc', info={'load_time': 9527}))
    pool.add_event(Event(2.0, 'wait', 'desc',info={'wait_time':0}))
    pool.add_event(Event(3.0, 'load', 'desc',info={'load_time':98}))
    pool.update_last_info('wait', {'wait_time': 10}, strict=False)
    print(pool.get_even_by_type('wait'))

    # test get_last_event
    print("==============test get_last_event================")
    event1 = pool.get_last_event('wait',strict=False)
    print(f"BEFORE: {event1}")
    event1.info['wait_time'] = 100000
    print(f"AFTER: {pool.get_last_event('wait',strict=False)}")

    # test get_event_by_time
    print("==============test get_event_by_time================")
    for t in range(100):
        pool.add_event(Event(time_stamp=t, event_type='test', desc='desc'))
    print(pool.get_event_by_time(21))


from __future__ import annotations
import time

class BaseDispatcher:
    def __init__(self):
        self.init_order_count = 0
        self.haul_order_count = 0
        self.back_order_count = 0
        self.init_order_time = 0
        self.haul_order_time = 0
        self.back_order_time = 0
        self.total_order_count = 0
        self.total_order_time = 0

    def __getattribute__(self, name):
        attr = object.__getattribute__(self, name)
        if name in ["give_init_order", "give_haul_order", "give_back_order"] and callable(attr):
            return self._track_calls_and_time_wrapper(attr, name)
        return attr

    def _track_calls_and_time_wrapper(self, method, method_type):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = method(*args, **kwargs)
            elapsed_time = (time.time() - start_time) * 1000

            if method_type == "give_init_order":
                self.init_order_count += 1
                self.init_order_time += elapsed_time
            elif method_type == "give_haul_order":
                self.haul_order_count += 1
                self.haul_order_time += elapsed_time
            elif method_type == "give_back_order":
                self.back_order_count += 1
                self.back_order_time += elapsed_time

            self.total_order_count = self.init_order_count + self.haul_order_count + self.back_order_count
            self.total_order_time = self.init_order_time + self.haul_order_time + self.back_order_time

            return result
        return wrapper

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        raise NotImplementedError("Subclass must implement this method.")

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        raise NotImplementedError("Subclass must implement this method.")

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        raise NotImplementedError("Subclass must implement this method.")

class TestDispatcher(BaseDispatcher):
    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        # 具体实现
        time.sleep(0.6)
        return 0

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        # 具体实现
        return 0

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # 具体实现
        return 0

if __name__ == "__main__":
    dispatcher = TestDispatcher()
    dispatcher.give_init_order("truck1", "mine1")
    dispatcher.give_init_order("truck1", "mine1")
    dispatcher.give_init_order("truck1", "mine1")
    dispatcher.give_init_order("truck1", "mine1")
    dispatcher.give_init_order("truck1", "mine1")
    dispatcher.give_haul_order("truck2", "mine2")
    dispatcher.give_back_order("truck3", "mine3")

    print("Init Order Count:", dispatcher.init_order_count)
    print("Haul Order Count:", dispatcher.haul_order_count)
    print("Back Order Count:", dispatcher.back_order_count)
    print("Total Order Count:", dispatcher.total_order_count)
    print("Init Order Time:", dispatcher.init_order_time)
    print("Haul Order Time:", dispatcher.haul_order_time)
    print("Back Order Time:", dispatcher.back_order_time)
    print("Total Order Time:", dispatcher.total_order_time)



"""
at the beginning of the operation, all the trucks is in a special parking lot called charging site
when a truck is ready to be loaded, it will be sent to the load site

this file is for the charging site
"""
from __future__ import annotations

class ChargingSite:
    def __init__(self, name:str,position:tuple=(0,0)):
        self.name = name
        self.position = position
        self.mine = None
        self.trucks = []

    def set_mine(self, mine:"Mine"):
        assert mine is not None, "mine can not be None"
        self.mine = mine

    def add_truck(self, truck:"Truck"):
        assert truck is not None, "truck can not be None"
        self.trucks.append(truck)

    def get_distance_to_load_site(self, load_site_index:int)->float:
        assert load_site_index >= 0, "load_site_index can not be negative"
        assert load_site_index < self.mine.road.load_site_num, "load_site_index can not be larger than load_site_num"
        return self.mine.road.charging_to_load[load_site_index]


import json
import matplotlib.pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from matplotlib.legend_handler import HandlerBase
import matplotlib.image as mpimg
import numpy as np
import matplotlib.animation as animation
import os

from tqdm import tqdm
from matplotlib.font_manager import FontProperties

# 定义字体属性
legend_font = FontProperties(family='Times New Roman', size=8)

plt.rcParams['font.family'] = 'PingFang HK'
plt.rcParams["axes.unicode_minus"] = False

class VisualGrapher:
    def __init__(self, json_path):
        with open(json_path, 'r', encoding='utf-8') as file:
            self.data = json.load(file)

        self.truck_colors = {}
        self.fig, self.ax = plt.subplots(figsize=(5, 5))

        # Use os.path.join to construct image paths relative to the script's location
        script_directory = os.path.dirname(__file__)
        self.img_mine = mpimg.imread(os.path.join(script_directory, 'materials', '矿山.png'))
        self.img_shovel = mpimg.imread(os.path.join(script_directory, 'materials', '挖机.png'))
        self.img_truck_hauling = mpimg.imread(os.path.join(script_directory, 'materials', 'truck_hauling.png'))
        self.img_truck_unhauling = mpimg.imread(os.path.join(script_directory, 'materials', 'truck_unhauling.png'))
        self.img_truck_initing = mpimg.imread(os.path.join(script_directory, 'materials', 'truck_initing.png'))
        self.img_dump = mpimg.imread(os.path.join(script_directory, 'materials', 'dump.png'))

    def update_progress(self, pbar):
        pbar.update()

    def create_animation(self, output_path='output.gif'):
        frames = len(self.data)
        with tqdm(total=frames, desc="Generating Animation") as pbar:
            ani = animation.FuncAnimation(self.fig, lambda i: self.plot_tick(i, pbar), frames=frames, repeat=False,
                                          blit=False)
            ani.save(output_path, writer='pillow', fps=10, dpi=300)

    def plot_tick(self, i, pbar):
        self.update_progress(pbar)
        try:
            tick_data = self.data[str(i)]
        except KeyError:
            return
        self.ax.clear()
        self.ax.set_xlim(0, 1)
        self.ax.set_ylim(0, 1)
        mine_data = tick_data['mine_states']
        # 1. 在图片上方展示矿山的总产量和服务次数
        self.ax.set_title(f"Production: {mine_data['produced_tons']:.2f} Tons, TruckCycles: {mine_data['service_count']} time: {i}/{len(self.data)-1}",fontproperties=legend_font)

        # self.ax.text(0.5, 0.95, f"Production: {mine_data['produced_tons']:.2f} Tons, TruckCycles: {mine_data['service_count']}",
        #              ha='center', va='center', fontsize=10, color='red')

        # 2. 准备动态图例的数据
        mine_stats = [
            ("Waiting Trucks", mine_data['waiting_truck_count']),
            ("Loading/Unloading Trucks", mine_data['load_unload_truck_count']),
            ("Moving Trucks", mine_data['moving_truck_count']),
            ("Repairing Trucks", mine_data['repairing_truck_count']),
            ("Road Jams", mine_data['road_jam_count']),
            ("Road Repairs", mine_data['road_repair_count']),
            ("Truck Repairs", mine_data['truck_repair']),
            ("Irreparable Trucks", mine_data['truck_unrepairable']),
            ("Random Events", mine_data['random_event_count'])
        ]

        # 3. 创建动态图例
        colors = ['red', 'blue', 'green', 'yellow', 'purple', 'orange', 'brown', 'pink', 'gray', 'cyan']
        legend_elements = [plt.Line2D([0], [0], marker='o', color='w', label=f'{name}: {value}',
                                      markerfacecolor=colors[i], markersize=5) for i, (name, value) in enumerate(mine_stats)]
        self.ax.legend(handles=legend_elements, loc='upper right', prop=legend_font)

        DUMP_SITE_SCALE = 0.1    # 可以调整卸载区的缩放比例
        LOAD_SITE_SCALE = 0.1
        SHOVEL_SCALE = 0.10
        TRUCK_SCALE = 0.06

        # 3. 绘制装载区并展示总吨数
        for load_site, load_data in tick_data['load_site_states'].items():
            self.place_image(load_data['position'], self.img_mine, zoom=LOAD_SITE_SCALE)
            self.ax.text(load_data['position'][0], load_data['position'][1] - 0.06, load_site,
                         ha='center', fontsize=4)
            stats_text = f"Tons: {load_data['tons']:.2f}," \
                            f"count: {load_data['service_count']}"
            self.ax.text(load_data['position'][0], load_data['position'][1] + 0.05, stats_text,
                            ha='center', fontsize=4)

        # 4. 绘制卸载区并展示总吨数
        for dump_site, dump_data in tick_data['dump_site_states'].items():
            self.place_image(dump_data['position'], self.img_dump, zoom=DUMP_SITE_SCALE)
            self.ax.text(dump_data['position'][0], dump_data['position'][1] - 0.05, dump_site,
                         ha='center', fontsize=4)
            stats_text = f"Tons: {dump_data['tons']:.2f}," \
                         f"count: {dump_data['service_count']}"
            self.ax.text(dump_data['position'][0], dump_data['position'][1] + 0.05, stats_text,
                         ha='center', fontsize=4)

        # 5. 绘制铲车并展示其产量和服务次数
        for shovel, shovel_data in tick_data['shovel_states'].items():
            self.place_image(shovel_data['position'], self.img_shovel, zoom=SHOVEL_SCALE)
            # 在图像底部展示铲车名称
            self.ax.text(shovel_data['position'][0], shovel_data['position'][1] - 0.04, shovel,
                         ha='center', va='bottom', fontsize=3)
            # 在图像顶部展示产量和服务次数
            stats_text = f"tons: {shovel_data['tons']:.2f}, count: {shovel_data['service_count']}"
            self.ax.text(shovel_data['position'][0], shovel_data['position'][1] + 0.04, stats_text,
                         ha='center', va='top', fontsize=4)

        # 绘制矿车
        for truck, truck_data in tick_data['truck_states'].items():
            if truck_data['state'] in [self.ON_CHARGING_SITE, self.ON_ROAD_TO_FIRST_LOAD]:
                img_truck = self.img_truck_initing
            elif truck_data['state'] == self.UNHAULING:
                img_truck = self.img_truck_unhauling
            elif truck_data['state'] == self.HAULING:
                img_truck = self.img_truck_hauling
            else:
                img_truck = self.img_truck_initing
            self.place_image(truck_data['position'], img_truck, zoom=TRUCK_SCALE)
            self.ax.text(truck_data['position'][0], truck_data['position'][1] - 0.035, truck, ha='center', fontsize=3,
                         color='black')


    def place_image(self, xy, img, zoom=1):
        """
        在指定的坐标上放置图像
        :param xy: 放置图像的坐标 (x, y)
        :param img: 图像数组
        :param zoom: 图像的缩放级别
        """
        im_offset = OffsetImage(img, zoom=zoom)
        ab = AnnotationBbox(im_offset, xy, frameon=False, box_alignment=(0.5, 0.5))
        self.ax.add_artist(ab)

# 状态常量定义
VisualGrapher.ON_CHARGING_SITE = -1
VisualGrapher.ON_ROAD_TO_FIRST_LOAD = -2
VisualGrapher.UNHAULING = 0
VisualGrapher.WAIT_FOR_LOAD = 1
VisualGrapher.LOADING = 2
VisualGrapher.HAULING = 3
VisualGrapher.WAIT_FOR_DUMP = 4
VisualGrapher.DUMPING = 5
VisualGrapher.REPAIRING = 6
VisualGrapher.UNREPAIRABLE = 7

class ImageHandler(HandlerBase):
    def __init__(self, image_ref, scale):
        super(ImageHandler, self).__init__()
        self.image_ref = image_ref
        self.scale = scale

    def create_artists(self, legend, orig_handle, xdescent, ydescent, width, height, fontsize, trans):
        zoom = self.scale / height  # 缩放图片以适应图例
        im = OffsetImage(self.image_ref, zoom=zoom)
        ab = AnnotationBbox(im, (width/2., height/2.), xycoords='axes fraction', frameon=False)
        return [ab]



if __name__ == '__main__':
    path = "/Users/mac/PycharmProjects/truck_shovel_mix/openmines_project/openmines/src/cli/results/MINE:NorthOpenPitMine_ALGO:RandomDispatcher_TIME:2024-01-21 14:26:04.gif"
    visualizer = VisualGrapher(path)
    visualizer.create_animation(output_path='output.gif')



import simpy

class ParkingLot:
    def __init__(self,name:str,position:tuple):
        self.name = name
        self.position = position
        self.queue_status = dict()


    def get_waiting_trucks(self, resource_list:list[simpy.Resource])->int:
        waiting_trucks = []
        for resource in resource_list:
            waiting_trucks += resource.queue
        return len(waiting_trucks)

    def monitor_resources(self, env, resources, res_objs, monitor_interval=1):
        """监控停车场对应所有资源的排队长度
            可以做到铲子、dumper单体级别的排队长度
        """
        self.queue_status["total"] = dict()
        while True:
            # 记录当前的排队长度
            all_queue_len = sum([len(resource.queue) for resource in resources])
            self.queue_status["total"][int(env.now)] = all_queue_len
            # 统计预期等待时间
            # 获取自定义请求包含的信息
            for index,res_obj in enumerate(res_objs):  # 遍历铲车或dumper对象
                res_name = res_obj.name
                load_time_list = []  # 当前装载点所有铲车的队列装载时间
                dump_time_list = []  # 当前卸载点所有dumper的队列卸载时间
                # 如果是铲车，那就对铲车队列等待时间进行统计
                if res_obj.type == "shovel":
                    shovel_name = res_obj.name
                    shovel_tons = res_obj.shovel_tons
                    shovel_cycle_time = res_obj.shovel_cycle_time
                    truck_load_time = 0
                    for i,load_request in enumerate(resources[index].queue):  # 遍历某铲车的装载请求
                        truck = load_request.truck
                        load_site = load_request.load_site
                        truck_load_time += (truck.truck_capacity / res_obj.shovel_tons) * res_obj.shovel_cycle_time
                    load_time_list.append(truck_load_time)  # 一个铲车的所有装载时间之和
                # 如果是dumper，那就对dumper队列等待时间进行统计
                if res_obj.type == "dumper":
                    dumper_name = res_obj.name
                    dumper_cycle_time = res_obj.dump_time
                    # 计算队列等待时间
                    dumper_queue_estimated_time = len(resources[index].queue)* dumper_cycle_time
                    dump_time_list.append(dumper_queue_estimated_time)
            if res_objs[-1].type == "shovel":
                EWA_LOAD_TIME = sum(load_time_list) / len(load_time_list)  # 装载区的平均等待时间
                load_site = res_objs[-1].load_site
                load_site.estimated_queue_wait_time = EWA_LOAD_TIME
            if res_objs[-1].type == "dumper":
                EWA_DUMP_TIME = sum(dump_time_list) / len(dump_time_list)  # dumpsite的平均卸载时间
                dump_site = res_objs[-1].dump_site
                dump_site.estimated_queue_wait_time = EWA_DUMP_TIME
            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)

    def monitor_resource(self, env, res_obj, resource, monitor_interval=1):
        """监控铲子、dumper单体级别的排队长度信息
            由于使用了自定义的LoadRequest/DumpRequest,
            所以request中会包含load_site/dumpsite,truck等信息

            res_obj: shovel or dumper
            resource: shovel or dumper resource
        """
        res_name = res_obj.name
        self.queue_status[res_name] = dict()
        while True:
            # 记录当前的排队长度
            self.queue_status[res_name][int(env.now)] = len(resource.queue)
            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)

class Shovel:
    def __init__(self, name:str, shovel_tons:float, shovel_cycle_time:float,position_offset:tuple=(0, 0.05)):
        self.name = name
        self.type = "shovel"
        self.load_site = None
        self.position = None
        self.position_offset = position_offset
        self.shovel_tons = shovel_tons  # the size of shovel bucket
        self.produced_tons = 0  # the produced tons of shovel
        self.service_count = 0  # the number of shovel-vehicle cycle
        self.shovel_cycle_time = shovel_cycle_time  # shovel-vehicle time took for one shovel of mine
        self.status = dict()  # the status of shovel

    def set_env(self, env:simpy.Environment):
        self.env = env
        self.res = simpy.Resource(env, capacity=1)

    def monitor_status(self, env, monitor_interval=1):
        """监控铲车的产量、服务次数等信息
        """
        while True:
            # 获取铲车信息
            self.status[int(env.now)] = {
                "produced_tons": self.produced_tons,
                "service_count": self.service_count,
            }
            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)


class LoadSite:
    def __init__(self, name:str, position:tuple):
        self.name = name
        self.position = position
        self.shovel_list = []
        self.parking_lot = None
        self.status = dict()  # the status of shovel
        self.produced_tons = 0  # the produced tons of shovel
        self.service_count = 0  # the number of shovel-vehicle cycle
        self.estimated_queue_wait_time = 0  # the estimation of total waiting time for coming trucks in queue
        self.load_site_productivity = 0  # the productivity of load site

    def set_env(self, env:simpy.Environment):
        self.env = env
        for shovel in self.shovel_list:
            shovel.set_env(env)

    def monitor_status(self, env, monitor_interval=1):
        """监控装载区的产量、服务次数等信息
        """
        while True:
            # 获取铲车信息
            for shovel in self.shovel_list:
                self.produced_tons += shovel.produced_tons
                self.service_count += shovel.service_count
            self.status[int(env.now)] = {
                "produced_tons": self.produced_tons,
                "service_count": self.service_count,
            }
            # 统计装载区的装载能力
            load_site_productivity = sum(
                shovel.shovel_tons / shovel.shovel_cycle_time for shovel in self.shovel_list)
            self.load_site_productivity = load_site_productivity
            # reset
            self.produced_tons = 0
            self.service_count = 0
            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)


    def show_shovels(self):
        shovel_names = []
        for shovel in self.shovel_list:
            shovel_names.append(shovel.name)
        print(f'{self.name} has {shovel_names}')

    def add_shovel(self, shovel:Shovel):
        # 铲子的位置是相对于卸载点的位置，和dumper的数目*偏移不同
        shovel.position = tuple(a + b for a, b in zip(self.position, shovel.position_offset))
        # 在shovel中添加自己的引用
        shovel.load_site = self
        self.shovel_list.append(shovel)

    def add_parkinglot(self, position_offset, name: str = None):
        if name is None:
            name = f'{self.name}_parking_lot'
        park_position = tuple(a + b for a, b in zip(self.position, position_offset))
        self.parking_lot = ParkingLot(name=name, position=park_position)

    def get_available_shovel(self)->Shovel:
        """
        这里是一个简单的贪心算法，返回第一个空闲的铲车
        TODO：下一个版本，也许可以将铲车暴露出来作为单个卸载区让其决策。目前仅贪心算法。
        :return:
        """
        for shovel in self.shovel_list:
            if shovel.res.count == 0:
                return shovel
        # 如果没有空闲的铲车 则返回最小队列的铲车
        min_queue_shovel = min(self.shovel_list, key=lambda shovel: len(shovel.res.queue))
        return min_queue_shovel


"""模拟的核心对象Mine

使用的时候需要先初始化Mine，
然后添加
LoadSite(Shovel)、DumpSite(Dumper)、
ChargingSite(Truck)，Road，Dispatcher等对象
最后调用Mine.run()方法开始仿真
"""
import os
import time

import simpy,logging,math
from functools import reduce

from openmines.src.charging_site import ChargingSite
from openmines.src.dispatcher import BaseDispatcher
from openmines.src.dump_site import DumpSite
from openmines.src.load_site import LoadSite
from openmines.src.road import Road
from openmines.src.truck import Truck
from openmines.src.utils.logger import MineLogger, LOG_FILE_PATH
from openmines.src.utils.ticker import TickGenerator
from openmines.src.utils.event import EventPool

class Mine:
    def __init__(self, name:str, log_path=LOG_FILE_PATH, log_file_level=logging.DEBUG, log_console_level=logging.INFO):
        self.env = simpy.Environment()
        self.name = name
        self.load_sites = []
        self.dump_sites = []
        self.trucks = []
        self.road = None
        self.dispatcher = None
        self.random_event_pool = EventPool()  # 随机事件池
        # summary
        self.produce_tons = 0  # the produced tons of this dump site
        self.service_count = 0  # the number of shovel-vehicle cycle in this dump site
        self.status = dict()  # the status of shovel
        # LOGGER配置
        print(log_path)
        self.global_logger = MineLogger(log_path=log_path, file_level=log_file_level, console_level=log_console_level)
        self.mine_logger = self.global_logger.get_logger(name)

    def monitor_status(self, env, monitor_interval=1):
        """监控卸载区的产量、服务次数等信息
            监控道路的状态
        """
        while True:
            """
            1.监控卸载区的产量、服务次数等信息
            """
            # 获取每个dumper信息并统计
            for dump_site in self.dump_sites:
                # 获取每个dumper信息并统计
                for dumper in dump_site.dumper_list:
                    self.produce_tons += dumper.dumper_tons
                    self.service_count += dumper.service_count
            # 获取卡车整体的统计信息
            working_truck_count = 0
            waiting_truck_count = 0
            load_unload_truck_count = 0
            moving_truck_count = 0
            repairing_truck_count = 0
            for truck in self.trucks:
                # 计算可工作的truck数量
                if truck.status != "unrepairable":
                    working_truck_count += 1
                # 计算正在排队中的truck数量
                if "waiting" in truck.status:
                    waiting_truck_count += 1
                # 计算正在装载卸载的truck数量
                if truck.status == "loading" or truck.status == "unloading":
                    load_unload_truck_count += 1
                # 计算moving的truck数量
                if truck.status == "moving":
                    moving_truck_count += 1
                # 计算正在维修的truck数量
                if truck.status == "repairing":
                    repairing_truck_count += 1
            # 计算出现的random event的数量
            road_jam_count = 0
            road_repair_count = 0
            truck_repair = 0
            truck_unrepairable = 0
            random_event_count = 0
            for event_key in self.random_event_pool.event_set.keys():
                event_type = self.random_event_pool.event_set[event_key].event_type
                if event_type == "RoadEvent:jam":
                    road_jam_count += 1
                if event_type == "RoadEvent:repair":
                    road_repair_count += 1
                if event_type == "TruckEvent:breakdown":
                    truck_repair += 1
                if event_type == "TruckEvent:unrepairable":
                    truck_unrepairable += 1
            random_event_count = road_jam_count + road_repair_count + truck_repair + truck_unrepairable
            self.status[int(env.now)] = {
                # KPIs
                "produced_tons": self.produce_tons,
                "service_count": self.service_count,
                # stats
                "working_truck_count": working_truck_count,
                "waiting_truck_count": waiting_truck_count,
                "load_unload_truck_count": load_unload_truck_count,
                "moving_truck_count": moving_truck_count,
                "repairing_truck_count": repairing_truck_count,
                # event stats
                "road_jam_count": road_jam_count,
                "road_repair_count": road_repair_count,
                "truck_repair": truck_repair,
                "truck_unrepairable": truck_unrepairable,
                "random_event_count":random_event_count
            }
            # reset
            self.produce_tons = 0
            self.service_count = 0
            """
            2.监控道路的状态
            """
            cur_time = env.now
            # 获取mine的路网
            road = self.road
            charging_to_load_road = road.charging_to_load
            road_matrix = road.road_matrix
            # 获取每条路的状态
            road_status = dict()
            ## 统计初始化init过程中的道路情况
            for i in range(self.road.load_site_num):
                charging_site_name = self.charging_site.name
                load_site_name = self.load_sites[i].name
                road_status[(charging_site_name, load_site_name)] = {"truck_jam_count": 0, "repair_count": 0,
                                                                 "truck_count": 0}
                # 统计道路上当前正在发生的随机事件个数、历史上发生的随机事件个数
                ## 获取jam事件
                jam_events = self.random_event_pool.get_even_by_type("RoadEvent:jam")
                for jam_event in jam_events:
                    """
                    info={"name": self.name, "status": "jam", "speed": 0,
                                                          "start_location": self.current_location.name,
                                                            "end_location": self.target_location.name,
                                                          "start_time": self.env.now, "est_end_time":
                    """
                    if jam_event.info["start_location"] == charging_site_name and jam_event.info[
                        "end_location"] == load_site_name \
                            and jam_event.info["start_time"] <= cur_time and jam_event.info["est_end_time"] >= cur_time:
                        road_status[(charging_site_name, load_site_name)]["truck_jam_count"] += 1
                ## 获取repair事件
                repair_events = self.random_event_pool.get_even_by_type("RoadEvent:repair")
                for repair_event in repair_events:
                    if repair_event.info["start_location"] == charging_site_name and repair_event.info[
                        "end_location"] == load_site_name \
                            and repair_event.info["repair_start_time"] <= cur_time and repair_event.info[
                        "repair_end_time"] >= cur_time:
                        road_status[(charging_site_name, load_site_name)]["repair_count"] += 1
                ## 统计道路上的truck
                for truck in self.trucks:
                    if truck.current_location == charging_site_name and truck.target_location == load_site_name:
                        road_status[(charging_site_name, load_site_name)]["truck_count"] += 1

            ## 统计haul过程中的道路情况
            for i in range(self.road.load_site_num):
                for j in range(self.road.dump_site_num):
                    load_site_name = self.load_sites[i].name
                    dump_site_name = self.dump_sites[j].name
                    road_status[(load_site_name, dump_site_name)] = {"truck_jam_count": 0, "repair_count": 0, "truck_count": 0}

                    # 统计道路上当前正在发生的随机事件个数、历史上发生的随机事件个数
                    ## 获取jam事件
                    jam_events = self.random_event_pool.get_even_by_type("RoadEvent:jam")
                    for jam_event in jam_events:
                        """
                        info={"name": self.name, "status": "jam", "speed": 0,
                                                              "start_location": self.current_location.name,
                                                                "end_location": self.target_location.name,
                                                              "start_time": self.env.now, "est_end_time":
                        """
                        if jam_event.info["start_location"] == load_site_name and jam_event.info["end_location"] == dump_site_name \
                            and jam_event.info["start_time"] <= cur_time and jam_event.info["est_end_time"] >= cur_time:
                            road_status[(load_site_name, dump_site_name)]["truck_jam_count"] += 1
                    ## 获取repair事件
                    repair_events = self.random_event_pool.get_even_by_type("RoadEvent:repair")
                    for repair_event in repair_events:
                        if repair_event.info["start_location"] == load_site_name and repair_event.info["end_location"] == dump_site_name \
                            and repair_event.info["repair_start_time"] <= cur_time and repair_event.info["repair_end_time"] >= cur_time:
                            road_status[(load_site_name, dump_site_name)]["repair_count"] += 1
                    ## 统计道路上的truck
                    for truck in self.trucks:
                        if truck.current_location == load_site_name and truck.target_location == dump_site_name:
                            road_status[(load_site_name, dump_site_name)]["truck_count"] += 1

            ## 统计unhaul过程中的道路情况
            for j in range(self.road.dump_site_num):
                for i in range(self.road.load_site_num):
                    load_site_name = self.load_sites[i].name
                    dump_site_name = self.dump_sites[j].name
                    road_status[(dump_site_name,load_site_name)] = {"truck_jam_count": 0, "repair_count": 0, "truck_count": 0}

                    # 统计道路上当前正在发生的随机事件个数、历史上发生的随机事件个数
                    ## 获取jam事件
                    jam_events = self.random_event_pool.get_even_by_type("RoadEvent:jam")
                    for jam_event in jam_events:
                        if jam_event.info["start_location"] == dump_site_name and jam_event.info["end_location"] == load_site_name \
                            and jam_event.info["start_time"] <= cur_time and jam_event.info["est_end_time"] >= cur_time:
                            road_status[(dump_site_name,load_site_name)]["truck_jam_count"] += 1
                    ## 获取repair事件
                    repair_events = self.random_event_pool.get_even_by_type("RoadEvent:repair")
                    for repair_event in repair_events:
                        if repair_event.info["start_location"] == dump_site_name and repair_event.info["end_location"] == load_site_name \
                            and repair_event.info["repair_start_time"] <= cur_time and repair_event.info["repair_end_time"] >= cur_time:
                            road_status[(dump_site_name,load_site_name)]["repair_count"] += 1
                    ## 统计道路上的truck
                    for truck in self.trucks:
                        if truck.current_location == dump_site_name and truck.target_location == load_site_name:
                            road_status[(dump_site_name,load_site_name)]["truck_count"] += 1

            self.road.road_status = road_status

            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)

    def add_load_site(self, load_site:LoadSite):
        load_site.set_env(self.env)
        self.load_sites.append(load_site)

    def add_charging_site(self, charging_site:ChargingSite):
        charging_site.set_mine(self)
        self.charging_site = charging_site
        self.trucks = charging_site.trucks
        for truck in self.trucks:
            truck.set_env(self)

    def add_dump_site(self, dump_site:DumpSite):
        dump_site.set_env(self.env)
        self.dump_sites.append(dump_site)

    def add_road(self, road:Road):
        assert road is not None, "road can not be None"
        road.set_env(self)
        self.road = road

    def add_dispatcher(self, dispatcher:BaseDispatcher):
        assert dispatcher is not None, "dispatcher can not be None"
        self.dispatcher = dispatcher

    def get_load_site_status(self):
        for load_site in self.load_sites:
            print(f'{load_site.name} has {len(load_site.res.users)} trucks')

    def get_dest_index_by_name(self,name:str):
        """
        使用名字去匹配铲子和卸载区 并返回其标号
        标号定义如下：
        None 没有匹配
        0～N-1 装载区标号
        N～N+M-1 卸载区标号

        :param name:
        :return:
        """
        assert name is not None, "name can not be None"
        # 查找装载区
        for i,loadsite in enumerate(self.load_sites):
            if loadsite.name == name:
                return i
        # 查找卸载区
        for i,dumpsite in enumerate(self.dump_sites):
            if dumpsite.name == name:
                return i+len(self.load_sites)
        return None

    def get_dest_obj_by_index(self,index:int):
        """
        使用标号去匹配铲子和卸载区 并返回其对象
        :param index:
        :return:
        """
        assert index is not None, "index can not be None"
        if index < len(self.load_sites):
            return self.load_sites[index]
        elif index < len(self.load_sites) + len(self.dump_sites):
            return self.dump_sites[index - len(self.load_sites)]
        else:
            return None

    def get_dest_obj_by_name(self,name:str):
        """
        使用名字去匹配装载区和卸载区 并返回其对象
        :param name:
        :return:
        """
        assert name is not None, "name can not be None"
        # 查找装载区
        for loadsite in self.load_sites:
            if loadsite.name == name:
                return loadsite
        # 查找卸载区
        for dumpsite in self.dump_sites:
            if dumpsite.name == name:
                return dumpsite
        # 查找充电站
        if self.charging_site.name == name:
            return self.charging_site
        return None

    def get_service_vehicle_by_name(self,name:str):
        """
        使用名字匹配铲子和dumper并返回其对象
        :param name:
        :return:
        """
        assert name is not None, "name can not be None"
        for loadsite in self.load_sites:
            for shovel in loadsite.shovel_list:
                if shovel.name == name:
                    return shovel
        for dumpsite in self.dump_sites:
            for dumper in dumpsite.dumper_list:
                if dumper.name == name:
                    return dumper
        return None


    def summary(self):
        """在仿真结束后，统计各个装载区和卸载区的产量、MatchingFactor等并记录在文件中
        目前统计的对象：
        1. 卸载区产量
        2. TruckCycleTime
        3. MatchingFactor
        4. TruckWaitTime、TotalWaitTime

        :return:
        """
        pass


    def start(self, total_time:float=60*8)->dict:
        """
        普通策略算法的仿真入口(包括RL策略推理时)
        :param total_time:
        :return:
        """
        assert self.road is not None, "road can not be None"
        assert self.dispatcher is not None, "dispatcher can not be None"
        assert self.charging_site is not None, "charging_site can not be None"
        assert len(self.load_sites) > 0, "load_sites can not be empty"
        assert len(self.dump_sites) > 0, "dump_sites can not be empty"
        assert len(self.trucks) > 0, "trucks can not be empty"
        assert total_time > 0, "total_time can not be negative"
        self.total_time = total_time
        self.mine_logger.info("simulation started")
        # start some monitor process for summary
        for load_site in self.load_sites:
            # 对停车场队列的监控
            self.env.process(load_site.parking_lot.monitor_resources(env=self.env,
                                                                    resources=[shovel.res for shovel in load_site.shovel_list],
                                                                     res_objs=load_site.shovel_list))
            for shovel in load_site.shovel_list:
                self.env.process(load_site.parking_lot.monitor_resource(env=self.env,res_obj=shovel,
                                                                        resource=shovel.res))
            # 对铲车产出的监控
            for shovel in load_site.shovel_list:
                self.env.process(shovel.monitor_status(env=self.env))
            # 对装载区产出、装载区队列的监控
            self.env.process(load_site.monitor_status(env=self.env))

        for dump_site in self.dump_sites:
            # 对停车场队列的监控
            self.env.process(dump_site.parking_lot.monitor_resources(env=self.env,
                                                                    resources=[dumper.res for dumper in dump_site.dumper_list],
                                                                     res_objs=dump_site.dumper_list))
            for dumper in dump_site.dumper_list:
                self.env.process(dump_site.parking_lot.monitor_resource(env=self.env,res_obj=dumper,
                                                                        resource=dumper.res))
            # 对卸载区产出、卸载区队列的监控
            self.env.process(dump_site.monitor_status(env=self.env))
            # 对dumper产出的监控
            for dumper in dump_site.dumper_list:
                self.env.process(dumper.monitor_status(env=self.env))
        # 对矿山整体监控
        self.env.process(self.monitor_status(env=self.env))
        # log in the truck as process
        for truck in self.trucks:
            self.env.process(truck.run())
        self.env.run(until=total_time)
        self.mine_logger.info("simulation finished")
        self.summary()
        ticks = self.dump_frames(total_time=total_time)
        return ticks

    def start_rl(self, total_time:float=60*8)->dict:
        """
        使用RL算法的仿真入口
        :param total_time:
        :return:
        """
        assert self.road is not None, "road can not be None"
        assert self.dispatcher is not None, "dispatcher can not be None"
        assert self.charging_site is not None, "charging_site can not be None"
        assert len(self.load_sites) > 0, "load_sites can not be empty"
        assert len(self.dump_sites) > 0, "dump_sites can not be empty"
        assert len(self.trucks) > 0, "trucks can not be empty"
        assert total_time > 0, "total_time can not be negative"
        self.total_time = total_time
        self.mine_logger.info("simulation started")
        # start some monitor process for summary
        for load_site in self.load_sites:
            # 对停车场队列的监控
            self.env.process(load_site.parking_lot.monitor_resources(env=self.env,
                                                                     resources=[shovel.res for shovel in
                                                                                load_site.shovel_list],
                                                                     res_objs=load_site.shovel_list))
            for shovel in load_site.shovel_list:
                self.env.process(load_site.parking_lot.monitor_resource(env=self.env, res_obj=shovel,
                                                                        resource=shovel.res))
            # 对铲车产出的监控
            for shovel in load_site.shovel_list:
                self.env.process(shovel.monitor_status(env=self.env))
            # 对装载区产出、装载区队列的监控
            self.env.process(load_site.monitor_status(env=self.env))

        for dump_site in self.dump_sites:
            # 对停车场队列的监控
            self.env.process(dump_site.parking_lot.monitor_resources(env=self.env,
                                                                     resources=[dumper.res for dumper in
                                                                                dump_site.dumper_list],
                                                                     res_objs=dump_site.dumper_list))
            for dumper in dump_site.dumper_list:
                self.env.process(dump_site.parking_lot.monitor_resource(env=self.env, res_obj=dumper,
                                                                        resource=dumper.res))
            # 对卸载区产出、卸载区队列的监控
            self.env.process(dump_site.monitor_status(env=self.env))
            # 对dumper产出的监控
            for dumper in dump_site.dumper_list:
                self.env.process(dumper.monitor_status(env=self.env))
        # 对矿山整体监控
        self.env.process(self.monitor_status(env=self.env))
        # log in the truck as process
        for truck in self.trucks:
            self.env.process(truck.run(is_rl_training=True))


    def dump_frames(self,total_time):
        """使用TickGenerator记录仿真过程中的数据
        将数据写入到文件中
        :return:
        """
        # tick generator
        self.tick_generator = TickGenerator(mine=self, tick_num=total_time)
        assert self.tick_generator is not None, "tick_generator can not be None"
        print("dumping frames...")
        self.tick_generator.run()
        # 获得年月日时分秒的字符串表示
        time_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        ticks = self.tick_generator.write_to_file(file_name=f'MINE:{self.name}_ALGO:{self.dispatcher.name}_TIME:{time_str}.json')
        return ticks


from __future__ import annotations

from typing import List

import numpy as np  # 导入NumPy库
import random,json,time
import openai


from openmines.src.dispatcher import BaseDispatcher
from openmines.src.load_site import LoadSite, Shovel
from openmines.src.dump_site import DumpSite, Dumper


class SPTFDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "SPTFDispatcher"
        self.np_random = np.random.RandomState()  # 创建NumPy的随机状态对象

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取其他卡车信息
        other_trucks = [each_truck for each_truck in mine.trucks if each_truck.name != truck.name]
        # 统计其他卡车的目的地并进行计数;过滤掉target_location为None的卡车;
        target_location_trucks = [each_truck for each_truck in other_trucks if each_truck.target_location is not None and isinstance(each_truck.target_location, LoadSite)]
        # 统计loadsite信息
        load_sites = mine.load_sites
        for _, each_load_site in enumerate(load_sites):
            each_load_site.coming_truck_list = []
            # 统计target_locations中本loadsite的数量
            for each_coming_truck in target_location_trucks:
                if each_coming_truck.target_location.name == each_load_site.name:
                    each_load_site.coming_truck_list.append(each_coming_truck)
            # processing time = coming truck processing time + estimated queue wait time
            each_load_site.processing_time_est = sum([each_truck.truck_capacity/each_load_site.load_site_productivity for each_truck in each_load_site.coming_truck_list]) + each_load_site.estimated_queue_wait_time

        # 计算index
        loadsite_index = np.argmin([loadsite.processing_time_est for loadsite in load_sites])

        return loadsite_index

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取其他卡车信息
        other_trucks = [each_truck for each_truck in mine.trucks if each_truck.name != truck.name]
        # 统计其他卡车的目的地并进行计数;过滤掉target_location为None的卡车;
        target_location_trucks = [each_truck for each_truck in other_trucks if truck.target_location is not None and isinstance(each_truck.target_location, DumpSite)]
        # 统计dumpsite信息
        dump_sites = mine.dump_sites
        for each_dump_site in dump_sites:
            each_dump_site.coming_truck_list: List["Truck"] = []
            for each_coming_truck in target_location_trucks:
                if each_coming_truck.target_location.name == each_dump_site.name:
                    each_dump_site.coming_truck_list.append(each_coming_truck)
            each_dump_site.processing_time_est = len(each_dump_site.coming_truck_list) + each_dump_site.estimated_queue_wait_time

        dumpsite_index = np.argmin([dumpsite.processing_time_est for dumpsite in dump_sites])
        return dumpsite_index

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # 这个方法可以与give_init_order方法完全一样，因为它们的逻辑是相同的
        return self.give_init_order(truck, mine)



if __name__ == "__main__":
    dispatcher = SPTFDispatcher()
    print(dispatcher.give_init_order(1,2))
    print(dispatcher.give_haul_order(1,2))
    print(dispatcher.give_back_order(1,2))

    print(dispatcher.total_order_count,dispatcher.init_order_count)


from __future__ import annotations
import numpy as np  # 导入NumPy库
import random,json,time
import openai


from openmines.src.dispatcher import BaseDispatcher
from openmines.src.load_site import LoadSite, Shovel
from openmines.src.dump_site import DumpSite, Dumper


class SQDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "SQDispatcher"
        self.np_random = np.random.RandomState()  # 创建NumPy的随机状态对象

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取其他卡车信息
        other_trucks = [each_truck for each_truck in mine.trucks if each_truck.name != truck.name]
        # 统计其他卡车的目的地并进行计数;过滤掉target_location为None的卡车;
        target_locations = [each_truck.target_location for each_truck in other_trucks if each_truck.target_location is not None and isinstance(each_truck.target_location, LoadSite)]
        # 统计loadsite信息
        load_sites = mine.load_sites
        for each_load_site in load_sites:
            each_load_site.coming_truck_num = 0
            # 统计target_locations中本loadsite的数量
            for each_target_location in target_locations:
                if each_target_location.name == each_load_site.name:
                    each_load_site.coming_truck_num += 1
        loadsite_queue_length = [load_site.parking_lot.queue_status["total"][int(mine.env.now)] for load_site in load_sites]
        # estimated_loadsite_queue_wait_times = [load_site.estimated_queue_wait_time for load_site in load_sites]
        # 选取coming_truck+queue_length最小的loadsite
        loadsite_index = np.argmin([loadsite.coming_truck_num + loadsite_queue_length[i] for i, loadsite in enumerate(load_sites)])


        # 获取dumpsite信息
        #avaliable_dumpsites = [dumpsite for dumpsite in mine.dump_sites if dumpsite.parking_lot is not None]
        #dump_site_names = [dumpsite.name for dumpsite in avaliable_dumpsites]
        #dumpsite_queue_length = [dumpsite.parking_lot.queue_status["total"][int(mine.env.now)] for dumpsite in
        #                        avaliable_dumpsites]
        #estimated_dumpsite_queue_wait_times = [dumpsite.estimated_queue_wait_time for dumpsite in avaliable_dumpsites]
        return loadsite_index

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        cur_location:LoadSite = truck.current_location
        # 获取其他卡车信息
        other_trucks = [each_truck for each_truck in mine.trucks if each_truck.name != truck.name]
        # 统计其他卡车的目的地并进行计数;过滤掉target_location为None的卡车;
        target_locations = [each_truck.target_location for each_truck in other_trucks if
                            each_truck.target_location is not None and isinstance(each_truck.target_location, DumpSite)]
        # 统计dumpsite信息
        dump_sites = mine.dump_sites
        for each_dump_site in dump_sites:
            each_dump_site.coming_truck_num = 0
            # 统计target_locations中本dumpsite的数量
            for each_target_location in target_locations:
                if each_target_location.name == each_dump_site.name:
                    each_dump_site.coming_truck_num += 1
        dumpsite_queue_length = [dump_site.parking_lot.queue_status["total"][int(mine.env.now)] for dump_site in
                                 dump_sites]
        # estimated_loadsite_queue_wait_times = [load_site.estimated_queue_wait_time for load_site in load_sites]
        # 选取coming_truck+queue_length最小的loadsite
        dumpsite_index = np.argmin(
            [dumpsite.coming_truck_num + dumpsite_queue_length[i] for i, dumpsite in enumerate(dump_sites)])

        # 获取dumpsite信息
        # avaliable_dumpsites = [dumpsite for dumpsite in mine.dump_sites if dumpsite.parking_lot is not None]
        # dump_site_names = [dumpsite.name for dumpsite in avaliable_dumpsites]
        # dumpsite_queue_length = [dumpsite.parking_lot.queue_status["total"][int(mine.env.now)] for dumpsite in
        #                        avaliable_dumpsites]
        # estimated_dumpsite_queue_wait_times = [dumpsite.estimated_queue_wait_time for dumpsite in avaliable_dumpsites]
        return dumpsite_index

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取其他卡车信息
        other_trucks = [each_truck for each_truck in mine.trucks if each_truck.name != truck.name]
        # 统计其他卡车的目的地并进行计数;过滤掉target_location为None的卡车;
        target_locations = [each_truck.target_location for each_truck in other_trucks if each_truck.target_location is not None and isinstance(each_truck.target_location, LoadSite)]
        # 统计loadsite信息
        load_sites = mine.load_sites
        for each_load_site in load_sites:
            each_load_site.coming_truck_num = 0
            # 统计target_locations中本loadsite的数量
            for each_target_location in target_locations:
                if each_target_location.name == each_load_site.name:
                    each_load_site.coming_truck_num += 1
        loadsite_queue_length = [load_site.parking_lot.queue_status["total"][int(mine.env.now)] for load_site in load_sites]
        # estimated_loadsite_queue_wait_times = [load_site.estimated_queue_wait_time for load_site in load_sites]
        # 选取coming_truck+queue_length最小的loadsite
        loadsite_index = np.argmin([loadsite.coming_truck_num + loadsite_queue_length[i] for i, loadsite in enumerate(load_sites)])


        # 获取dumpsite信息
        #avaliable_dumpsites = [dumpsite for dumpsite in mine.dump_sites if dumpsite.parking_lot is not None]
        #dump_site_names = [dumpsite.name for dumpsite in avaliable_dumpsites]
        #dumpsite_queue_length = [dumpsite.parking_lot.queue_status["total"][int(mine.env.now)] for dumpsite in
        #                        avaliable_dumpsites]
        #estimated_dumpsite_queue_wait_times = [dumpsite.estimated_queue_wait_time for dumpsite in avaliable_dumpsites]
        return loadsite_index



if __name__ == "__main__":
    dispatcher = SQDispatcher()
    print(dispatcher.give_init_order(1,2))
    print(dispatcher.give_haul_order(1,2))
    print(dispatcher.give_back_order(1,2))

    print(dispatcher.total_order_count,dispatcher.init_order_count)


import json
import os
import pandas as pd
from matplotlib import pyplot as plt
from pathlib import Path

from matplotlib.font_manager import FontProperties

# 设置Times New Roman字体
times_new_roman = FontProperties(family='Times New Roman', size=8)

class Charter:
    def __init__(self, config_file):
        self.config_file = config_file
        self.config_name = Path(config_file).name
        self.img_name = self.config_name.replace('.json', '.tiff')
        self.tab_name = self.config_name.replace('.json', '') + "_table" + '.tiff'
        self.fig_img = None
        self.fig_tab = None

        # 使用Path来处理路径，以确保跨平台兼容性
        self.result_path = Path.cwd() / 'results'

        # 确保结果目录存在
        self.result_path.mkdir(exist_ok=True)

        # 使用Path对象组合路径
        self.img_path = self.result_path / self.img_name
        self.tab_path = self.result_path / self.tab_name

    def draw(self, states_dict):
        # 绘制图像
        fig_img, axs = plt.subplots(2, 1, figsize=(3.5, 5))  # 创建两个子图的图像

        # 存储最终的production数据
        final_table = []
        # 遍历每个调度器的结果并绘图
        for dispatcher_name, states in states_dict.items():
            # 获取必要统计信息
            decision_latency = states['summary']['avg_time_per_order']
            matching_factor = states['summary']['MatchingFactor']
            total_wait_time = states['summary']['TotalWaitTime']
            road_jams = states['summary']['RoadJams']
            produced_tons = states['produced_tons_list'][-1]

            # 绘制产量曲线
            axs[0].plot(states['times'], states['produced_tons_list'], label=dispatcher_name, linewidth=0.7)
            # 绘制waitingTruck曲线
            axs[1].plot(states['times'], states['waiting_truck_count_list'], label=dispatcher_name, linewidth=0.7)
            # final_table
            final_table.append({
                'Name': dispatcher_name,
                'Produced Tons': f"{round(produced_tons, 2):.2f}",
                'Decision Latency (ms)': f"{round(decision_latency, 1):.1f}",
                'Matching Factor': f"{round(matching_factor, 2):.2f}",
                'Total Wait Time': f"{round(total_wait_time, 2):.2f}",
                "Road Jams": road_jams
            })
        # 设置坐标轴刻度的字体大小
        axs[0].tick_params(axis='both', which='major', labelsize=6)
        axs[1].tick_params(axis='both', which='major', labelsize=6)

        # 设置图像的图例和标题
        axs[0].legend(prop={'size': 5}, loc='upper left')
        axs[1].legend(prop={'size': 5}, loc='center right')
        axs[0].set_title("Production Over Time", fontproperties=times_new_roman)
        axs[1].set_title("Waiting Trucks Over Time", fontproperties=times_new_roman)
        axs[0].set_xlabel("Time (mins)", fontproperties=times_new_roman)
        axs[0].set_ylabel("Production (tons)", fontproperties=times_new_roman)
        axs[1].set_xlabel("Time (mins)", fontproperties=times_new_roman)
        axs[1].set_ylabel("Number of Waiting Trucks", fontproperties=times_new_roman)

        plt.tight_layout()
        self.fig_img = fig_img


        # 绘制表格
        fig_tab, ax_table = plt.subplots(figsize=(15, 5))  # 创建单独的Figure用于表格
        df = pd.DataFrame(final_table)
        ax_table.axis('tight')
        ax_table.axis('off')
        table = ax_table.table(cellText=df.values, colLabels=df.columns, loc='center')
        table.auto_set_font_size(False)
        table.set_fontsize(10)
        table.scale(1, 1.5)

        plt.tight_layout()
        self.fig_tab = fig_tab

    def save(self):
        # 保存图像
        if self.fig_img is None or self.fig_tab is None:
            raise Exception("You must draw the chart first!")
        self.fig_img.savefig(self.img_path, dpi=300, format='tiff')
        self.fig_tab.savefig(self.tab_path, dpi=300, format='tiff')



from setuptools import setup, find_packages

import openmines

# 读取requirements.txt文件中的内容
with open('openmines/requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='openmines',
    version=openmines.__version__,
    author='stone91',  # 添加作者
    description='Mining Fleet Dispatch Algorithm Framework',  # 添加描述（英文）
    license='MIT',  # 添加许可证
    packages=find_packages(),
    package_data={
        # 确保你的包名正确
        'openmines': ['./src/utils/visualization/materials/*'],
    },
    entry_points={
        'console_scripts': [
            'openmines=openmines.src.cli.run:main',
        ],
    },
    install_requires=required,  # 使用requirements.txt中的依赖
    # 可以在此添加更多元数据，如作者邮箱、项目网址等
)


import simpy

from openmines.src.load_site import ParkingLot


class Dumper:
    def __init__(self, name:str, dumper_cycle_time:float, position_offset:tuple=(0, 0.01)):
        self.name = name
        self.type = "dumper"
        self.dump_site = None
        self.position = None
        self.position_offset = position_offset
        self.dumper_tons:float = 0.0  #  the tons of material the dumper has received
        self.service_count = 0
        self.dump_time = dumper_cycle_time  # shovel-vehicle time took for one shovel of mine
        self.status = dict()  # the status of shovel

    def set_env(self, env:simpy.Environment):
        self.env = env
        self.res = simpy.Resource(env, capacity=1)

    def monitor_status(self, env, monitor_interval=1):
        """监控卸载位的产量、服务次数等信息
        """
        while True:
            self.status[int(env.now)] = {
                "produced_tons": self.dumper_tons,
                "service_count": self.service_count,
            }
            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)

class DumpSite:
    def __init__(self, name:str, position:tuple):
        self.name = name
        self.position = position
        self.dumper_list = []
        self.parking_lot = None
        self.tons = 0  # 卸载点的总吨数 用于统计
        self.truck_visits = 0  # 卸载点的总车次数 用于统计
        self.status = dict()  # the status of shovel
        self.produce_tons = 0  # the produced tons of this dump site
        self.service_count = 0  # the number of shovel-vehicle cycle in this dump site
        self.estimated_queue_wait_time = 0  # the estimation of total waiting time for coming trucks in queue

    def set_env(self, env:simpy.Environment):
        self.env = env
        for dumper in self.dumper_list:
            dumper.set_env(env)

    def monitor_status(self, env, monitor_interval=1):
        """监控卸载区的产量、服务次数等信息
        """
        while True:
            # 获取每个dumper信息并统计
            for dumper in self.dumper_list:
                self.produce_tons += dumper.dumper_tons
                self.service_count += dumper.service_count
            self.status[int(env.now)] = {
                "produced_tons": self.produce_tons,
                "service_count": self.service_count,
            }
            # reset
            self.produce_tons = 0
            self.service_count = 0
            # 等待下一个监控时间点
            yield env.timeout(monitor_interval)

    def add_dumper(self, dumper:Dumper):
        dumper_count = len(self.dumper_list)
        dumper.position = tuple(a + b for a, b in zip(self.position, dumper.position_offset*(dumper_count+1)))
        # 在dumper中添加自己引用
        dumper.dump_site = self
        self.dumper_list.append(dumper)

    def get_produce_tons(self):
        self.tons = sum([dumper.dumper_tons for dumper in self.dumper_list])
        return self.tons

    def show_dumpers(self):
        dumper_names = []
        for dumper in self.dumper_list:
            dumper_names.append(dumper.name)
        print(f'{self.name} has {dumper_names}')

    def add_parkinglot(self, position_offset, name:str=None):
        if name is None:
            name = f'{self.name}_parking_lot'
        park_position = tuple(a + b for a, b in zip(self.position, position_offset))
        self.parking_lot = ParkingLot(name=name, position=park_position)


    def get_available_dumper(self)->Dumper:
        """
        这里是一个简单的贪心算法，返回第一个空闲的卸载点
        TODO：下一个版本，也许可以将铲车暴露出来作为单个卸载区让其决策。目前仅贪心算法。
        :return:
        """
        for dumper in self.dumper_list:
            if dumper.res.count == 0:
                return dumper
        # 如果没有空闲的铲车 则返回最小队列的铲车
        min_queue_dumper = min(self.dumper_list, key=lambda dumper: len(dumper.res.queue))
        return min_queue_dumper


from __future__ import annotations

import time

import numpy as np

from openmines.src.dispatcher import BaseDispatcher
from openmines.src.dump_site import DumpSite
from openmines.src.load_site import LoadSite


class NearestDispatcher(BaseDispatcher):
    def __init__(self):
        super().__init__()
        self.name = "NearestDispatcher"

    def give_init_order(self, truck: "Truck", mine: "Mine") -> int:
        charging_to_load = mine.road.charging_to_load
        # 获取最短数值的index
        min_index = charging_to_load.index(min(charging_to_load))
        return min_index

    def give_haul_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取当前位置
        current_location = truck.current_location
        assert isinstance(current_location, LoadSite), "current_location is not a LoadSite"
        # 获取当前位置的index
        cur_index = mine.load_sites.index(current_location)
        # 获取当前位置到所有dump site的距离
        cur_to_dump:np.array = mine.road.road_matrix[cur_index,:]
        # 获取nparray最短数值的index
        min_index = cur_to_dump.argmin()
        # 从第一个load site开始
        return min_index

    def give_back_order(self, truck: "Truck", mine: "Mine") -> int:
        # 获取当前位置
        current_location = truck.current_location
        assert isinstance(current_location, DumpSite), "current_location is not a DumpSite"
        # 获取当前位置的index
        cur_index = mine.dump_sites.index(current_location)
        # 获取当前位置到所有load site的距离
        cur_to_load = mine.road.road_matrix[:,cur_index]
        # 获取nparray最短数值的index
        min_index = cur_to_load.argmin()
        return min_index

if __name__ == "__main__":
    dispatcher = NearestDispatcher()
    print(dispatcher.give_init_order(1,2))
    print(dispatcher.give_init_order(1, 2))
    print(dispatcher.give_init_order(1, 2))
    print(dispatcher.give_init_order(1, 2))
    print(dispatcher.give_haul_order(1,2))
    print(dispatcher.give_back_order(1,2))

    print(dispatcher.total_order_count,dispatcher.init_order_count,dispatcher.init_order_time,dispatcher.total_order_time)
